import typing
from functools import wraps
from starlette.routing import Route
from red_warden.auth import auth_from_bearer_token, auth_from_session, required_scope
from red_warden.sessions import session_from_cookie
from red_warden.i18n import RWMemoryI18n
from red_warden.responses import RWJSONResponse


class RWRoutingException(Exception):
    pass


def api(_func=None):
    def decorator_api(func):
        @wraps(func)
        async def wrapper_api(request):
            request.state.api = True
            return RWJSONResponse(await func(request))

        return wrapper_api

    if _func is None:
        return decorator_api
    else:
        return decorator_api(_func)


class RWApiRoute(Route):
    def __init__(
        self,
        path: str,
        endpoint: typing.Callable,
        scopes=None,
        methods: typing.List[str] = None,
        include_in_schema: bool = True,
    ) -> None:
        name = "%s_%s" % (endpoint.__self__.__name__, endpoint.__name__)
        required_scope[name] = scopes

        super().__init__(
            path,
            api(auth_from_bearer_token(endpoint)),
            methods=methods,
            name=name,
            include_in_schema=include_in_schema,
        )


class RWRoute(Route):
    def __init__(
        self,
        path: str,
        endpoint: typing.Callable,
        scope=None,
        methods: typing.List[str] = None,
        include_in_schema: bool = True,
        name=None,
    ) -> None:

        if not name:
            name = "%s_%s" % (
                endpoint.__self__.__name__,
                endpoint.__name__,
            )

        required_scope[name] = scope

        super().__init__(
            path,
            session_from_cookie(auth_from_session(endpoint)),
            methods=methods,
            name=name,
            include_in_schema=include_in_schema,
        )

    @staticmethod
    def translate(endpoint, methods=None):
        routes = []
        route_name = "%s_%s" % (
            endpoint.__self__.__name__,
            endpoint.__name__,
        )

        translated = RWMemoryI18n.all(route_name)
        for locale, route_path in translated.items():
            if not route_path.startswith("/"):
                raise RWRoutingException(
                    "Check route path translations for route %s" % route_name
                )

            routes.append(
                RWRoute(
                    route_path,
                    endpoint,
                    name="%s_%s" % (route_name, locale),
                    methods=methods,
                )
            )
        return routes



import os
import json
from starlette.middleware.base import BaseHTTPMiddleware


class RWI18n:
    _locales = []
    _locale = None

    @property
    def locale(self):
        if not self._locales:
            raise Exception("No locales defined")

        return self._locale

    @locale.setter
    def locale(self, value):
        if value in self._locales:
            self._locale = value
        else:
            raise Exception("Missing locale %s" % value)

    @classmethod
    def load_from_json_file(cls, json_file, locale):
        with open(json_file, mode="r", encoding="utf8") as f:
            translations = json.load(f)
            cls._add_translations(translations, locale)

        if locale not in cls._locales:
            cls._locales.append(locale)

    @classmethod
    def load_from_kv_file(cls, kv_file, locale):
        with open(kv_file, mode="r", encoding="utf8") as f:
            for line in f:
                line = line.strip()
                if line:
                    key, message = line.split("=")
                    cls._add_translation(key.strip(), message.strip(), locale)

        if locale not in cls._locales:
            cls._locales.append(locale)

    @classmethod
    def load_from_kv_directory(cls, kv_directory):
        for file in os.listdir(kv_directory):
            if file.endswith(".txt"):
                cls.load_from_kv_file(
                    "%s/%s" % (kv_directory, file), file.replace(".txt", "")
                )

    def t(
        self,
        key,
        locale=None,
        missing_placeholder="___MISSING_TRANSLATION_%s___",
        **kwargs
    ):
        if not locale:
            locale = self._locale

        if (locale not in self._locales) or not self._has_translation(key, locale):
            return missing_placeholder % key

        return self._get_translation(key, locale, **kwargs)

    @classmethod
    def all(cls, key):
        translations = {}
        for locale in cls._locales:
            translations[locale] = cls._get_translation(key, locale)
        return translations

    @classmethod
    def _has_translation(cls, key, locale):
        raise NotImplementedError

    @classmethod
    def _get_translation(cls, key, locale, **kwargs):
        raise NotImplementedError

    @classmethod
    def _add_translation(cls, key, message, locale):
        raise NotImplementedError

    @classmethod
    def _add_translations(cls, translations, locale):
        raise NotImplementedError


class RWMemoryI18n(RWI18n):
    _translations = {}

    @classmethod
    def __check_locale(cls, locale):
        if locale not in cls._translations.keys():
            cls._translations[locale] = {}

    @classmethod
    def _has_translation(cls, key, locale):
        cls.__check_locale(locale)
        return key in cls._translations[locale]

    @classmethod
    def _get_translation(cls, key, locale, **kwargs):
        cls.__check_locale(locale)
        translated = cls._translations[locale].get(key, key)
        return translated.format(**kwargs) if kwargs else translated

    @classmethod
    def _add_translation(cls, key, message, locale):
        cls.__check_locale(locale)
        cls._translations[locale][key] = message

    @classmethod
    def _add_translations(cls, locale, translations):
        cls.__check_locale(locale)
        cls._translations[locale].update(translations)


class RWI18nMiddleware(BaseHTTPMiddleware):
    _adapter_class = None

    def __init__(self, app, adapter_class):
        self._adapter_class = adapter_class
        super().__init__(app)

    async def dispatch(self, request, call_next):
        i18n = self._adapter_class()
        request.state.i18n = i18n

        try:
            locale = (
                str(request.url).replace(str(request.base_url), "").split("/", 1)[0]
            )
            i18n.locale = locale
        except:
            pass

        return await call_next(request)



import os
import logging
from random import choice
from starlette.config import Config as StarletteConfig
from starlette.datastructures import CommaSeparatedStrings, Secret
from starlette.middleware import Middleware
from enum import Enum


class _Config(dict):
    def __init__(self):
        super().__init__()

        # the Environment Vars are prioritary, after it tries to load the ".env" file
        self._cfg = StarletteConfig(env_file=os.path.join(os.getcwd(), ".env"))

        # --- APP Config ---
        # APP_NAME must be unique for all the application modules
        self.env_string("RW_NAME")
        self.env_string("RW_BACKEND_ENDPOINTS")
        self.env_string("RW_BACKEND_USERNAME")
        self.env_string("RW_BACKEND_PASSWORD")
        self.env_string("RW_BACKEND_DB_NAME")
        self.env_string("RW_BACKEND_PARAMS", "{}")
        self.env_secret("RW_SIGNER_SECRET")
        self.env_int("RW_SIGNER_TIMEOUT", 60)

        self.env_string("RW_PATH", default=os.getcwd())
        self.env_string("RW_ENV", default="production")
        self.env_bool("RW_DEBUG")
        self.env_string("RW_LOG_LEVEL", default="INFO")
        self.env_string(
            "RW_LOG_FORMAT",
            default=(
                "%(levelname) -8s %(asctime)s [ %(lineno) -5d] %(pathname) -50s: %(message)s"
            ),
        )  # %(thread)d for thread ID

        # --- VIEWS Config
        self["RW_VIEWS_TEMPLATE_DIR"] = os.getcwd() + "/views"

    def env_string(self, name, default=None):
        self[name] = self._cfg(name, default=default)

    def env_string_list(self, name, default=None):
        self[name] = self._cfg(name, cast=CommaSeparatedStrings, default=default or [])

    def env_bool(self, name, default=False):
        self[name] = self._cfg(name, cast=bool, default=default)

    def env_int(self, name, default=0):
        self[name] = self._cfg(name, cast=int, default=default)

    def env_secret(self, name):
        self[name] = self._cfg(name, cast=Secret, default=None)


RWConfig = _Config()

Logger = logging.getLogger(RWConfig["RW_NAME"])
Logger.setLevel(RWConfig["RW_LOG_LEVEL"])
_logger_handler = logging.StreamHandler()
Logger.addHandler(_logger_handler)
_logger_handler.setFormatter(logging.Formatter(RWConfig["RW_LOG_FORMAT"]))


if not RWConfig["RW_SIGNER_SECRET"]:
    raise Exception(
        """
Missing SIGNER secret key (env: RW_SIGNER_SECRET). Try with one of these:
%s
%s
%s
        """
        % (
            os.urandom(16).hex(),
            os.urandom(16).hex(),
            os.urandom(16).hex(),
        )
    )


class RedWardens:
    _red_wardens = {}

    @classmethod
    def get_endpoint(cls, red_warden_module_name, raise_if_missing=True):
        rw = cls._red_wardens.get(red_warden_module_name, None)
        if not rw:
            if raise_if_missing:
                raise Exception(
                    "RedWarden module %s does not exist" % red_warden_module_name
                )
            return None

        return choice(rw)

    @classmethod
    def add(cls, red_warden_module_name, endpoints):
        if not endpoints:
            raise Exception(
                "No endpoints specified for RedWarden module %s"
                % red_warden_module_name
            )
        if isinstance(endpoints, str):
            endpoints = endpoints.split(",")
        elif not isinstance(endpoints, list):
            endpoints = [endpoints]

        rw = cls._red_wardens.get(red_warden_module_name, None)
        if rw:
            raise Exception(
                "RedWarden module %s already exists" % red_warden_module_name
            )

        cls._red_wardens[red_warden_module_name] = endpoints


class Engines:
    _engines = {}

    @classmethod
    def get_all(cls):
        return cls._engines

    @classmethod
    def get(cls, engine_name, raise_if_missing=True):
        engine = cls._engines.get(engine_name, None)
        if raise_if_missing and not engine:
            raise Exception("Engine %s does not exist" % engine_name)
        return engine

    @classmethod
    def add(cls, engine_name, engine):
        cls._engines[engine_name] = engine


class Middlewares:
    _middlewares = []

    @classmethod
    def get_all(cls):
        return cls._middlewares

    @classmethod
    def insert(cls, position, middleware, **options):
        cls._middlewares.insert(position, Middleware(middleware, **options))

    @classmethod
    def add(cls, middleware, **options):
        cls._middlewares.append(Middleware(middleware, **options))



import graphene
from red_warden.graphql import (
    RWGraphQLNode,
    RWGraphQLObjectType,
    RWGraphQLYesNoField,
    RWGraphQLConnection,
)
from red_warden.datazones.RWManager import KeyModel


class Key(RWGraphQLObjectType):
    """A key used for authentication"""

    @classmethod
    async def check_permissions(cls, info, node):
        return node

    model = KeyModel

    class Meta:
        interfaces = (RWGraphQLNode,)

    login = graphene.String(description="The key name")
    enabled = RWGraphQLYesNoField(description="Is key enabled?")
    mfa_enabled = RWGraphQLYesNoField(
        description="Is multifactor authentication enabled?"
    )
    # tenants = RWGraphQLConnection.create_field_from_import(
    #     "schemas.link_tenant_key.connections.KeyTenantsConnection.all"
    # )



from enum import Enum


class RWMultiTenancyEnum(str, Enum):
    GLOBAL = "global"
    TENANT = "tenant"


class RWBackendTypeEnum(str, Enum):
    MYSQL = "mysql"
    MONGODB = "mongodb"
    REDIS = "redis"


class RWBackendDestinationEnum(str, Enum):
    GLOBAL = "global"
    TENANTS_AVAILABLE = "tenants_available"
    TENANTS_FULL = "tenants_full"


class RWYesNoEnum(str, Enum):
    YES = "Y"
    NO = "N"



import json
from typing import Optional
from aioauth.types import RequestMethod
from aioauth.requests import (
    Query,
    Post,
    Request as OAuth2Request,
)
from aioauth.config import Settings
from aioauth.collections import HTTPHeaderDict
from starlette.routing import Mount
from starlette.exceptions import HTTPException
from starlette.responses import Response
from starlette.status import HTTP_401_UNAUTHORIZED, HTTP_404_NOT_FOUND
from red_warden.config import RWConfig
from red_warden.datalayer import RWDataLayer
from red_warden.helpers import import_string
from red_warden.mvc import (
    RWValidationModelBase,
    RWRestController,
    RWController,
)
from red_warden.oauth2 import RWOAuth2Server
from red_warden.routing import RWApiRoute, RWRoute
from .models import (
    RWBackendModel,
    RWTenantModel,
    RWKeyModel,
    RWSessionModel,
    RWOAuth2ClientModel,
    RWOAuth2AuthorizationCodeModel,
    RWOAuth2TokenModel,
    RWAclRoleModel,
    RWAclPermissionModel,
    RWAclRolePermissionLinkModel,
    RWAclRoleKeyLinkModel,
)
from .logic import RWSessionLogic


class RWBackendController(RWRestController):
    model = RWBackendModel

    @classmethod
    def get_routes(cls):
        return [
            Mount(
                "/backends",
                routes=[
                    RWApiRoute("/", cls.load_all, ["manager"]),
                    RWApiRoute("/{id}", cls.load_by_id, ["manager"]),
                    RWApiRoute("/", cls.create, ["manager"], methods=["POST"]),
                    RWApiRoute("/{id}", cls.update, ["manager"], methods=["PUT"]),
                    RWApiRoute("/{id}", cls.delete, ["manager"], methods=["DELETE"]),
                    RWApiRoute(
                        "/upgrade-schemas",
                        cls.upgrade_schemas,
                        [],
                        methods=["POST"],
                    ),
                ],
            )
        ]

    @classmethod
    async def upgrade_schemas(cls, request):
        class ValidationModel(RWValidationModelBase):
            tenant_id: Optional[str]

        data = await request.form()
        data = ValidationModel(**data).dict()
        return {
            "data": await RWDataLayer.upgrade_other_datapaths_migrations(
                tenant_id=data.get("tenant_id", None)
            )
        }

    @classmethod
    async def create(cls, request):
        results = await super().create(request)
        await RWDataLayer.reload_backends()
        return results

    @classmethod
    async def update(cls, request):
        results = await super().update(request)
        await RWDataLayer.reload_backends()
        return results

    @classmethod
    async def delete(cls, request):
        results = await super().delete(request)
        await RWDataLayer.reload_backends()
        return results


class RWTenantController(RWRestController):
    model = RWTenantModel

    @classmethod
    def get_routes(cls):
        return [
            Mount(
                "/tenants",
                routes=[
                    RWApiRoute("/", cls.load_all, ["manager"]),
                    RWApiRoute("/{id}", cls.load_by_id, ["manager"]),
                    RWApiRoute("/", cls.create, ["manager"], methods=["POST"]),
                    RWApiRoute("/{id}", cls.update, ["manager"], methods=["PUT"]),
                    RWApiRoute("/{id}", cls.delete, ["manager"], methods=["DELETE"]),
                ],
            )
        ]


class RWKeyController(RWRestController):
    model = RWKeyModel

    @classmethod
    def get_routes(cls):
        return [
            Mount(
                "/keys",
                routes=[
                    RWApiRoute("/", cls.load_all, ["manager"]),
                    RWApiRoute("/{id}", cls.load_by_id, ["manager"]),
                    RWApiRoute("/", cls.create, ["manager"], methods=["POST"]),
                    RWApiRoute("/{id}", cls.update, ["manager"], methods=["PUT"]),
                    RWApiRoute("/{id}", cls.delete, ["manager"], methods=["DELETE"]),
                    RWApiRoute(
                        "/authenticate",
                        cls.authenticate,
                        ["manager"],
                        methods=["POST"],
                    ),
                    RWApiRoute(
                        "/{id}/mfa",
                        cls.toggle_mfa,
                        ["manager"],
                        methods=["POST"],
                    ),
                    RWApiRoute(
                        "/{id}/otp",
                        cls.check_otp,
                        ["manager"],
                        methods=["POST"],
                    ),
                ],
            )
        ]

    @classmethod
    async def check_otp(cls, request):
        data = await request.form()
        data["key_id"] = request.path_params.get("id", None)
        result = RWSessionLogic.check_otp(**data)

        if result is None:
            raise HTTPException(HTTP_404_NOT_FOUND)
        else:
            return result

    @classmethod
    async def toggle_mfa(cls, request):
        data = await request.form()
        data["key_id"] = request.path_params.get("id", None)
        result = RWSessionLogic.toggle_mfa(**data)

        if result is None:
            raise HTTPException(HTTP_404_NOT_FOUND)
        else:
            return result

    @classmethod
    async def authenticate(cls, request):
        """
        Used to re-check the password for an already logged in user.
        This is not the "login" method, as this is an API that should be asked
        before dangerous operations are made in frontend.

        This is not the "login" method also because it's not asking for MFA.
        """
        result = await RWSessionLogic.authenticate(**await request.form())

        if result is None:
            raise HTTPException(HTTP_404_NOT_FOUND)
        elif not result:
            raise HTTPException(HTTP_401_UNAUTHORIZED)
        else:
            return result


class RWSessionController(RWRestController):
    model = RWSessionModel

    @classmethod
    def get_routes(cls):
        return [
            Mount(
                "/sessions",
                routes=[
                    RWApiRoute("/", cls.load_all, ["manager"]),
                    RWApiRoute("/{id}", cls.load_by_id, ["manager"]),
                    RWApiRoute("/", cls.create, ["manager"], methods=["POST"]),
                    RWApiRoute("/{id}", cls.update, ["manager"], methods=["PUT"]),
                    RWApiRoute("/{id}", cls.delete, ["manager"], methods=["DELETE"]),
                ],
            )
        ]


class RWOAuth2ClientController(RWRestController):
    model = RWOAuth2ClientModel

    @classmethod
    def get_routes(cls):
        return [
            Mount(
                "/oauth2-clients",
                routes=[
                    RWApiRoute("/", cls.load_all, ["manager"]),
                    RWApiRoute("/{id}", cls.load_by_id, ["manager"]),
                    RWApiRoute("/", cls.create, ["manager"], methods=["POST"]),
                    RWApiRoute("/{id}", cls.update, ["manager"], methods=["PUT"]),
                    RWApiRoute("/{id}", cls.delete, ["manager"], methods=["DELETE"]),
                ],
            )
        ]


class RWOAuth2AuthorizationCodeController(RWRestController):
    model = RWOAuth2AuthorizationCodeModel

    @classmethod
    def get_routes(cls):
        return [
            Mount(
                "/oauth2-authorization-codes",
                routes=[
                    RWApiRoute("/", cls.load_all, ["manager"]),
                    RWApiRoute("/{id}", cls.load_by_id, ["manager"]),
                    RWApiRoute("/", cls.create, ["manager"], methods=["POST"]),
                    RWApiRoute("/{id}", cls.update, ["manager"], methods=["PUT"]),
                    RWApiRoute("/{id}", cls.delete, ["manager"], methods=["DELETE"]),
                ],
            )
        ]


class RWOAuth2TokenController(RWRestController):
    model = RWOAuth2TokenModel

    @classmethod
    def get_routes(cls):
        return [
            Mount(
                "/oauth2-tokens",
                routes=[
                    RWApiRoute("/", cls.load_all, ["manager"]),
                    RWApiRoute("/{id}", cls.load_by_id, ["manager"]),
                    RWApiRoute("/", cls.create, ["manager"], methods=["POST"]),
                    RWApiRoute("/{id}", cls.update, ["manager"], methods=["PUT"]),
                    RWApiRoute("/{id}", cls.delete, ["manager"], methods=["DELETE"]),
                ],
            )
        ]


class RWOAuth2FlowController(RWController):
    settings = Settings(INSECURE_TRANSPORT=RWConfig["RW_ENV"] == "local")

    @classmethod
    def get_routes(cls):
        return [
            RWRoute("/oauth2/authorize", cls.authorize, scope=["logged_in"]),
            RWRoute("/oauth2/token", cls.token, methods=["POST"]),
        ]

    @classmethod
    async def authorize(cls, request):
        """
        Endpoint to interact with the resource owner and obtain an authorization grant.

        See Section 4.1.1: https://tools.ietf.org/html/rfc6749#section-4.1.1
        """
        return await cls._to_starlette_response(
            await RWOAuth2Server.create_authorization_response(
                await cls._to_oauth2_request(request)
            )
        )

    @classmethod
    async def token(cls, request):
        """
        Endpoint to obtain an access and/or ID token by presenting an authorization grant or refresh token.

        See Section 4.1.3: https://tools.ietf.org/html/rfc6749#section-4.1.3
        """
        oauth2_request = await cls._to_oauth2_request(request)
        oauth2_response = await RWOAuth2Server.create_token_response(oauth2_request)
        return await cls._to_starlette_response(oauth2_response)

    @classmethod
    async def _to_oauth2_request(cls, request):
        """Converts Starlette Request instance to OAuth2Request instance"""
        form = await request.form()

        post = dict(form)
        query_params = dict(request.query_params)
        method = request.method
        headers = HTTPHeaderDict(**request.headers)
        url = str(request.url)
        return OAuth2Request(
            settings=cls.settings,
            method=RequestMethod[method],
            headers=headers,
            post=Post(**post),
            query=Query(**query_params),
            url=url,
            user=request.state.identity.get("key_id", None),
        )

    @staticmethod
    async def _to_starlette_response(oauth2_response):
        """Converts OAuth2Response instance to Starlette Response instance"""
        response_content = (
            oauth2_response.content if oauth2_response.content is not None else {}
        )
        headers = dict(oauth2_response.headers)
        status_code = oauth2_response.status_code
        content = json.dumps(response_content)

        return Response(content=content, headers=headers, status_code=status_code)


class RWAclRoleController(RWRestController):
    model = RWAclRoleModel
    role_permission_link_model = import_string(
        "red_warden.datazones.RWManager.RWAclRolePermissionLinkModel"
    )
    # TODO role_key_link_model = import_string("red_warden.datazones.RWManager.RWAclRoleKeyLinkModel")

    @classmethod
    def get_routes(cls):
        return [
            Mount(
                "/acl-roles",
                routes=[
                    RWApiRoute("/", cls.load_all, ["manager"]),
                    RWApiRoute("/{id}", cls.load_by_id, ["manager"]),
                    RWApiRoute("/", cls.create, ["manager"], methods=["POST"]),
                    RWApiRoute("/{id}", cls.update, ["manager"], methods=["PUT"]),
                    RWApiRoute("/{id}", cls.delete, ["manager"], methods=["DELETE"]),
                    RWApiRoute(
                        "/{id}/permissions", cls.load_role_permissions, ["manager"]
                    ),
                    RWApiRoute(
                        "/{id}/permissions/{permission_id}",
                        cls.create_role_permission,
                        ["manager"],
                        methods=["POST"],
                    ),
                    RWApiRoute(
                        "/{id}/permissions/{permission_id}",
                        cls.delete_role_permission,
                        ["manager"],
                        methods=["DELETE"],
                    ),
                ],
            )
        ]

    @classmethod
    async def load_role_permissions(cls, request):
        model_id = request.path_params.get("id", None)

        pagination = {
            "after": request.query_params.get("after", None),
            "before": request.query_params.get("before", None),
            "limit": request.query_params.get("limit", 0),
        }

        sort = request.query_params.get("sort", None)
        if sort:
            sort = sort.split(",")

        filters = request.query_params.get("filters", None)
        if filters:
            filters = filters.split(",")
        else:
            filters = []
        filters.append("acl_role_id:eq:%s" % model_id)

        data, pagination = await RWDataLayer.load_all(
            cls.role_permission_link_model,
            request.state.identity,
            pagination,
            sort,
            filters,
        )

        results = {"data": data}
        if pagination:
            results["pagination"] = pagination

        return results

    @classmethod
    async def _load_role_permission(cls, request, acl_role_id, acl_permission_id):
        return await RWDataLayer.load_one(
            cls.role_permission_link_model,
            request.state.identity,
            filters=[
                "acl_role_id:eq:%s" % acl_role_id,
                "acl_permission_id:eq:%s" % acl_permission_id,
            ],
        )

    @classmethod
    async def create_role_permission(cls, request):
        acl_role_id = request.path_params.get("id", None)
        acl_permission_id = request.path_params.get("permission_id", None)

        role_permission = await cls._load_role_permission(
            request, acl_role_id, acl_permission_id
        )
        if role_permission:
            await RWDataLayer.remove(
                cls.role_permission_link_model,
                role_permission["id"],
                request.state.identity,
            )

        body = {
            "acl_role_id": acl_role_id,
            "acl_permission_id": acl_permission_id,
        }
        return {
            "data": await RWDataLayer.save(
                cls.role_permission_link_model, None, body, request.state.identity
            )
        }

    @classmethod
    async def delete_role_permission(cls, request):
        acl_role_id = request.path_params.get("id", None)
        acl_permission_id = request.path_params.get("permission_id", None)

        role_permission = await cls._load_role_permission(
            request, acl_role_id, acl_permission_id
        )
        if role_permission:
            data = await RWDataLayer.remove(
                cls.role_permission_link_model,
                role_permission["id"],
                request.state.identity,
            )
            return {"data": data}


class RWAclPermissionController(RWRestController):
    model = RWAclPermissionModel

    @classmethod
    def get_routes(cls):
        return [
            Mount(
                "/acl-permissions",
                routes=[
                    RWApiRoute("/", cls.load_all, ["manager"]),
                    RWApiRoute("/{id}", cls.load_by_id, ["manager"]),
                    RWApiRoute("/", cls.create, ["manager"], methods=["POST"]),
                    RWApiRoute("/{id}", cls.update, ["manager"], methods=["PUT"]),
                    RWApiRoute("/{id}", cls.delete, ["manager"], methods=["DELETE"]),
                ],
            )
        ]



import graphene
from red_warden.graphql import RWGraphQLNode, RWGraphQLYesNoField
from red_warden.datazones.RWManager import TenantModel
from red_warden.datazones.graphql.tenant.types import Tenant


class CreateTenant(graphene.ClientIDMutation):
    tenant = graphene.Field(Tenant, required=True)

    class Input:
        parent_id = graphene.String()
        name = graphene.String(required=True)
        password = graphene.String(required=True)
        params = graphene.JSONString(required=False)
        enabled = RWGraphQLYesNoField(required=True)

    @classmethod
    async def mutate_and_get_payload(cls, root, info, **kwargs):
        db = TenantModel.get_datazone_db(info)

        async with db.transaction():
            kwargs["id"] = await db.get_uuid()

            parent_id = kwargs.pop("parent_id", None)
            if parent_id:
                node = await RWGraphQLNode.get_node_from_global_id(
                    info, parent_id, "Tenant", ignore_permissions=True
                )
                kwargs["parent_id"] = node.id

            await db.execute(TenantModel.insert().values(kwargs))
            kwargs.pop("parent_id", None)
            return cls(tenant=Tenant(**kwargs))



import graphene
from red_warden.graphql import (
    RWGraphQLConnection,
    RWGraphQLStringFilter,
    RWGraphQLYesNoField,
    RWGraphQLNode,
    RWGraphQLObjectType,
)
from red_warden.datazones.RWManager.models import BackendModel

# TODO from red_warden.datazones.graphql.datapath.connections import DatapathConnection


class BackendTypeField(graphene.Enum):
    Mysql = "mysql"
    MongoDB = "mongodb"
    Redis = "redis"


class Backend(RWGraphQLObjectType):
    @classmethod
    async def check_permissions(cls, info, node):
        return node

    model = BackendModel

    class Meta:
        interfaces = (RWGraphQLNode,)

    name = graphene.String(description="The backend name")
    type = BackendTypeField(description="The backend type")
    params = graphene.JSONString(description="A dictionary of parameters")
    endpoints = graphene.String(description="A comma separated list of endpoints")
    enabled = RWGraphQLYesNoField(description="Is backend enabled?")
    # TODO datapaths = DatapathConnection.create_field("all_by_backend")


class CreateBackend(graphene.ClientIDMutation):
    backend = graphene.Field(Backend, required=True)

    class Input:
        name = graphene.String(required=True)
        type = BackendTypeField(required=True)
        enabled = RWGraphQLYesNoField(required=True)
        endpoints = graphene.List(graphene.String, required=True)
        params = graphene.JSONString()

    @classmethod
    async def mutate_and_get_payload(cls, root, info, **kwargs):
        db = BackendModel.get_datazone_db(info)

        async with db.transaction():
            kwargs["id"] = await db.get_uuid()

            params = kwargs.get("params", None)
            if not params:
                kwargs["params"] = {}

            await db.execute(BackendModel.insert().values(kwargs))
            return cls(backend=Backend(**kwargs))


class BackendConnection(RWGraphQLConnection):
    class Meta:
        node = Backend
        filters = {
            "default": {
                "name": RWGraphQLStringFilter(),
                "type": BackendTypeField(),
                "enabled": RWGraphQLYesNoField(),
                "endpoints": RWGraphQLStringFilter(),
            }
        }

    @staticmethod
    def resolve_all(_, info, **args):
        return BackendModel.select()



import json
from datetime import datetime
from time import time
from typing import Optional
from aioauth.server import AuthorizationServer
from aioauth.storage import BaseStorage
from aioauth.models import Token, AuthorizationCode, Client
from cryptography.fernet import Fernet
from red_warden.config import RWConfig
from red_warden.enums import RWYesNoEnum
from red_warden.helpers import import_string
from red_warden.datalayer import RWDataLayer
from red_warden.datazones.RWManager.logic import RWSessionLogic


class _Storage(BaseStorage):
    def __init__(self):
        self.fernet = Fernet(str(RWConfig["RW_DATABASE_SECRET"]))

    async def get_client(
        self, request, client_id: str, client_secret: Optional[str] = None
    ) -> Optional[Client]:
        """Get client record from the database by provided request from user.

        Returns:
            `Client` instance if client exists in db.
            `None` if no client in db.
        """
        model = import_string("red_warden.datazones.RWManager.RWOAuth2ClientModel")
        client = await RWDataLayer.load_one(model, filters=["name:eq:%s" % client_id])
        if not client:
            return None

        if client_secret:
            # TODO verify the client secret!!
            pass

        return Client(
            client_id=client["id"],
            client_secret=client["secret"],
            grant_types=client["grant_types"].split(","),
            response_types=client["response_types"].split(","),
            redirect_uris=client["redirect_uris"] or [],
            scope=client["scope"] or "",
        )

    async def create_authorization_code(
        self,
        request,
        client_id: str,
        scope: str,
        response_type: str,
        redirect_uri: str,
        code_challenge_method: Optional[str],
        code_challenge: Optional[str],
    ) -> AuthorizationCode:
        """Generates an authorization token and stores it in the database.
        Warning:
            Generated authorization token *must* be stored in the database.
        Note:
            This must is used by the response type
            :py:class:`aioauth.respose_type.ResponseTypeAuthorizationCode`.
        Args:
            request: An :py:class:`aioauth.requests.Request`.
            client_id: A user client ID.
            scope: The scopes for the token.
            response_type: An :py:class:`aioauth.types.ResponseType`.
            redirect_uri: The redirect URI.
            code_challenge_method: An :py:class:`aioauth.types.CodeChallengeMethod`.
            code_challenge: Code challenge string.
        Returns:
            An :py:class:`aioauth.models.AuthorizationCode` object.
        """

        # generate an authorization token
        code = self.fernet.encrypt(
            json.dumps({"key_id": request.user, "scope": scope}).encode()
        ).decode("utf-8")

        authorization_code = AuthorizationCode(
            code=code,
            client_id=client_id,
            redirect_uri=redirect_uri,
            response_type=response_type,
            scope=scope,
            auth_time=int(time()),
            code_challenge_method=code_challenge_method,
            code_challenge=code_challenge,
            expires_in=request.settings.AUTHORIZATION_CODE_EXPIRES_IN,
        )

        # saves the Authorization Code to the db
        model = import_string(
            "red_warden.datazones.RWManager.RWOAuth2AuthorizationCodeModel"
        )
        await RWDataLayer.save(
            model,
            None,
            {
                "code": authorization_code.code,
                "oauth2_client_id": authorization_code.client_id,
                "key_id": request.user,
                "redirect_uri": authorization_code.redirect_uri,
                "response_type": authorization_code.response_type,
                "scope": authorization_code.scope,
                "auth_time": datetime.utcfromtimestamp(authorization_code.auth_time),
                "expires_in": authorization_code.expires_in,
                "nonce": authorization_code.nonce,
                "code_challenge": authorization_code.code_challenge,
                "code_challenge_method": authorization_code.code_challenge_method,
            },
        )
        return authorization_code

    @staticmethod
    async def _get_authorization_code_from_db(client_id, code):
        model = import_string(
            "red_warden.datazones.RWManager.RWOAuth2AuthorizationCodeModel"
        )
        auth_code = await RWDataLayer.load_one(
            model, filters=["code:eq:%s" % code, "oauth2_client_id:eq:%s" % client_id]
        )
        if not auth_code:
            return None

        return auth_code

    async def get_authorization_code(
        self, request, client_id: str, code: str
    ) -> Optional[AuthorizationCode]:

        auth_code = await self._get_authorization_code_from_db(client_id, code)
        if not auth_code:
            return None

        return AuthorizationCode(
            code=auth_code["code"],
            client_id=auth_code["oauth2_client_id"],
            redirect_uri=auth_code["redirect_uri"],
            response_type=auth_code["response_type"],
            scope=auth_code["scope"],
            auth_time=auth_code["auth_time"].timestamp(),
            expires_in=auth_code["expires_in"],
            code_challenge=auth_code["code_challenge"],
            code_challenge_method=auth_code["code_challenge_method"],
            nonce=auth_code["nonce"],
        )

    async def delete_authorization_code(
        self, request, client_id: str, code: str
    ) -> None:
        auth_code = await self._get_authorization_code_from_db(client_id, code)
        if auth_code:
            model = import_string(
                "red_warden.datazones.RWManager.RWOAuth2AuthorizationCodeModel"
            )
            await RWDataLayer.remove(model, auth_code["id"])

    async def create_token(self, request, client_id: str, scope: str) -> Token:
        token = await super().create_token(request, client_id, scope)
        print("scope", scope)

        # saves the Token to the db
        if request.post.code:
            data = self.fernet.decrypt(request.post.code.encode("utf-8")).decode(
                "utf-8"
            )
            data = json.loads(data)
            key_id = data["key_id"]
            # scope = data["scope"]
        else:
            model = import_string("red_warden.datazones.RWManager.RWKeyModel")
            key = await RWDataLayer.load_one(
                model, filters=["login:eq:%s" % request.post.username]
            )
            key_id = key["id"]
            # scope = scope

        model = import_string("red_warden.datazones.RWManager.RWOAuth2TokenModel")
        await RWDataLayer.save(
            model,
            None,
            {
                "access_token": token.access_token,
                "refresh_token": token.refresh_token,
                "scope": token.scope,
                "issued_at": datetime.utcfromtimestamp(token.issued_at),
                "access_token_expires_in": token.expires_in,
                "refresh_token_expires_in": token.refresh_token_expires_in,
                "oauth2_client_id": token.client_id,
                "key_id": key_id,
                "token_type": token.token_type,
                "revoked": RWYesNoEnum.YES if token.revoked else RWYesNoEnum.NO,
            },
        )

        return token

    async def get_token(
        self,
        request,
        client_id: str,
        access_token: Optional[str] = None,
        refresh_token: Optional[str] = None,
    ) -> Optional[Token]:
        model = import_string("red_warden.datazones.RWManager.RWOAuth2TokenModel")

        filters = ["oauth2_client_id:eq:%s" % client_id]
        if access_token:
            filters.append("access_token:eq:%s" % access_token)
        elif refresh_token:
            filters.append("refresh_token:eq:%s" % refresh_token)
        else:
            raise Exception("Please specify an access token or a refresh token")

        token = await RWDataLayer.load_one(model, filters=filters)
        if not token:
            return None

        return Token(
            access_token=token["access_token"],
            refresh_token=token["refresh_token"],
            scope=token["scope"],
            issued_at=token["issued_at"].timestamp(),
            expires_in=token["access_token_expires_in"],
            refresh_token_expires_in=token["refresh_token_expires_in"],
            client_id=token["oauth2_client_id"],
            token_type=token["token_type"],
            revoked=True if token["revoked"] == RWYesNoEnum.YES else False,
            user=token["key_id"],
        )

    async def revoke_token(self, request, token):
        """Revokes an existing token. The `revoked`

        Flag of the Token must be set to True
        """
        token_record = ...
        token_record.revoked = True
        token_record.save()

    async def authenticate(self, request) -> bool:
        if await RWSessionLogic.authenticate(
            username=request.post.username, password=request.post.password
        ):
            return True


RWOAuth2Server = AuthorizationServer(storage=_Storage())


#
# class Flow:
#
#     @classmethod
#
#
#     @staticmethod
#     async def introspect(request):
#         """Endpoint returns information about a token.
#
#         See Section 2.1: https://tools.ietf.org/html/rfc7662#section-2.1
#         """
#         # oauth2_request: OAuth2Request = await to_oauth2_request(request)
#         # oauth2_response: OAuth2Response = await server.create_token_introspection_response(
#         #     oauth2_request
#         # )
#         #
#         # return await to_fastapi_response(oauth2_response)
#         pass



import sys
import json
import aiofiles
import datetime
import uuid
import socket
import random
import traceback
from red_warden.config import RWConfig
from pydantic import ValidationError
from importlib import import_module


def get_my_ip():
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        # doesn't even have to be reachable
        s.connect(("10.255.255.255", 1))
        the_ip = s.getsockname()[0]
    except:
        the_ip = "127.0.0.1"
    finally:
        s.close()
    return the_ip


def get_random_element(elements, separator=","):
    if isinstance(elements, str):
        elements = elements.split(separator)

    return random.choice(elements)


class RWCustomJsonEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, uuid.UUID):
            return obj.hex.upper()
        elif isinstance(obj, datetime.datetime):
            return datetime.datetime.strftime(obj, "%Y%m%d%H%M%S")
        elif isinstance(obj, datetime.date):
            return datetime.date.strftime(obj, "%Y%m%d")
        elif isinstance(obj, datetime.time):
            return datetime.time.strftime(obj, "%H%M%S")
        elif isinstance(obj, bytes):
            return obj.hex().upper()

        return super().default(obj)


def exception_as_dict(ex):
    if isinstance(ex, ValidationError):
        msg = ex.errors()
    else:
        msg = str(ex)

    if msg:
        error = {"message": msg}
    else:
        error = {}

    if RWConfig["RW_DEBUG"]:
        exception_type, exception_object, exception_traceback = sys.exc_info()
        error.update(
            {"type": exception_type.__name__, "trace": traceback.format_stack()}
        )
    return error


# taken from Graphene
def import_string(dotted_path, dotted_attributes=None):
    """
    Import a dotted module path and return the attribute/class designated by the
    last name in the path. When a dotted attribute path is also provided, the
    dotted attribute path would be applied to the attribute/class retrieved from
    the first step, and return the corresponding value designated by the
    attribute path. Raise ImportError if the import failed.
    """
    try:
        module_path, class_name = dotted_path.rsplit(".", 1)
    except ValueError:
        raise ImportError("%s doesn't look like a module path" % dotted_path)

    module = import_module(module_path)

    try:
        result = getattr(module, class_name)
    except AttributeError:
        raise ImportError(
            'Module "%s" does not define a "%s" attribute/class'
            % (module_path, class_name)
        )

    if not dotted_attributes:
        return result
    else:
        attributes = dotted_attributes.split(".")
        traveled_attributes = []
        try:
            for attribute in attributes:
                traveled_attributes.append(attribute)
                result = getattr(result, attribute)
            return result
        except AttributeError:
            raise ImportError(
                'Module "%s" does not define a "%s" attribute inside attribute/class "%s"'
                % (module_path, ".".join(traveled_attributes), class_name)
            )


class ViteHelper:
    """
    This helper prepares the tags to be added when developing along Vite.

    from red_warden.helpers import ViteHelper
    from red_warden.mvc import RWController
    RWController.add_custom_view_data(
        await ViteHelper.get_js_and_css('/app/example_app/manifest.json', 'main.js')
    )
    """

    manifest = None

    @classmethod
    def _js_tag(cls, entry):
        url = (
            cls._asset_url(entry)
            if RWConfig["RW_ENV"] == "production"
            else "http://localhost:3000/%s" % entry
        )

        return (
            '<script type="module" crossorigin src="%s"></script>' % url if url else ""
        )

    @classmethod
    def _css_tag(cls, entry):
        tags = ""
        if RWConfig["RW_ENV"] == "production":
            for url in cls._css_urls(entry):
                tags += '<link rel="stylesheet" href="%s">' % url
        return tags

    @classmethod
    def _asset_url(cls, entry):
        item = cls.manifest.get(entry, None)
        return item.get("file", "") if item else ""

    @classmethod
    def _css_urls(cls, entry):
        urls = []
        item = cls.manifest.get(entry, None)

        if item:
            for css_file in item.get("css", []):
                urls.append("/dist/%s" % css_file)

        return urls

    @classmethod
    def _js_preload_imports(cls, entry):
        results = ""
        if RWConfig["RW_ENV"] == "production":
            for url in cls._imports_urls(entry):
                results += '<link rel="modulepreload" href="%s">' % url
        return results

    @classmethod
    def _imports_urls(cls, entry):
        urls = []
        manifest_entry = cls.manifest.get(entry, None)
        if manifest_entry:
            for _import in manifest_entry.get("imports", []):
                urls.append("/dist/%s" % _import)
        return urls

    @classmethod
    async def get_js_and_css(cls, manifest_file, entry):
        async with aiofiles.open(manifest_file, mode="r") as f:
            contents = await f.read()
        cls.manifest = json.loads(contents)

        return {
            "js": cls._js_tag(entry),
            "preload": cls._js_preload_imports(entry),
            "cls": cls._css_tag(entry),
        }



import graphene
from red_warden.graphql import (
    RWGraphQLNode,
    RWGraphQLConnection,
    RWGraphQLObjectType,
)
from models import DatapathModel, BackendModel

# --- Nodes -------------------------------------------------------------------
class Datapath(RWGraphQLObjectType):
    model = DatapathModel

    class Meta:
        interfaces = (RWGraphQLNode,)

    backend = graphene.Field("schemas.backend.types.Backend")
    name = graphene.String(description="The datapath name")
    params = graphene.JSONString(description="A dictionary of parameters")
    tenants = RWGraphQLConnection.create_field_from_import(
        "schemas.link_tenant_datapath.connections.DatapathTenantsConnection.all"
    )

    @staticmethod
    async def resolve_backend(parent, info, **args):
        db = BackendModel.get_datazone_db(info)
        return await db.fetch_one(
            BackendModel.select().where(BackendModel.c.id == parent.backend_id)
        )



import asyncio
import uuid
import base64
import json
from abc import ABC
from aioredis import ConnectionForcedCloseError, ConnectionClosedError
from red_warden.config import RWConfig, Backends, Logger


class RWEngine(ABC):
    _shutdown_event = asyncio.Event()

    async def run(self, *args, **kwargs):
        raise NotImplementedError

    async def stop(self):
        self._shutdown_event.set()


class RWRedisRpcEngine(RWEngine):
    _redis_rpc = None
    _redis_pubsub = None

    def __init__(self, redis_backend_name):
        self._redis_backend_name = redis_backend_name
        self._rpc_key = "%s:rpc" % RWConfig["RW_APP_NAME"].lower()
        self._pubsub_key = "%s:pubsub" % RWConfig["RW_APP_NAME"].lower()

    async def _rpc_listener(self):
        while not self._shutdown_event.is_set():
            try:
                self._redis_rpc = Backends.create_instance(self._redis_backend_name)
                await self._redis_rpc.connect()

                while not self._shutdown_event.is_set():
                    _, value = await self._redis_rpc.blpop(
                        "%s:requests" % self._rpc_key
                    )
                    data = json.loads(base64.b64decode(value).decode())

                    if data["rpc_name"] not in ["run", "stop", "execute"]:
                        rpc = getattr(self, data["rpc_name"], None)
                        if rpc:
                            resp = rpc(*data["args"], **data["kwargs"])
                            value = base64.b64encode(json.dumps(resp).encode())
                            resp_key = "%s:responses:%s" % (
                                self._rpc_key,
                                data["token"],
                            )
                            await self._redis_rpc.lpush(resp_key, value)
                            await self._redis_rpc.expire(resp_key, 10)

            except (
                ConnectionForcedCloseError,
                ConnectionClosedError,
                ConnectionRefusedError,
            ):
                if not self._shutdown_event.is_set():
                    Logger.info("Reconnecting RPC listener to Redis backend...")
                    await asyncio.sleep(0.5)

    async def _pubsub_listener(self):
        while not self._shutdown_event.is_set():
            try:
                self._redis_pubsub = Backends.create_instance(self._redis_backend_name)
                await self._redis_pubsub.connect()

                for ch in await self._redis_pubsub.subscribe(self._pubsub_key):
                    async for message in ch.iter():
                        print("TODO! Got message:", message)
            except (
                ConnectionForcedCloseError,
                ConnectionClosedError,
                ConnectionRefusedError,
            ):
                if not self._shutdown_event.is_set():
                    Logger.info("Reconnecting PubSub listener to Redis backend...")
                    await asyncio.sleep(0.5)

    async def run(self, *args, **kwargs):
        t1 = asyncio.create_task(self._rpc_listener())
        t2 = asyncio.create_task(self._pubsub_listener())
        await asyncio.gather(*[t1, t2])

    async def stop(self):
        await super().stop()
        if self._redis_rpc:
            await self._redis_rpc.disconnect()
        if self._redis_pubsub:
            await self._redis_pubsub.disconnect()

    async def execute(self, rpc_name, *args, **kwargs):
        rc = Backends.create_instance(self._redis_backend_name)
        await rc.connect()

        token = str(uuid.uuid4)
        req = {
            "token": token,
            "rpc_name": rpc_name,
            "args": args,
            "kwargs": kwargs,
        }
        data = base64.b64encode(json.dumps(req).encode())
        req_key = "%s:requests" % self._rpc_key
        await rc.lpush(req_key, data)
        await rc.expire(req_key, 10)

        resp = await rc.blpop("%s:responses:%s" % (self._rpc_key, token), timeout=5)
        if not resp:
            raise Exception("Redis request timed out")
        data = json.loads(base64.b64decode(resp[1]).decode())
        await rc.disconnect()
        _redis_client = None
        return data


"""
from red_warden.helpers import get_random_element, CustomJsonEncoder
import traceback
import functools
import pika
import json
from pika.adapters.asyncio_connection import AsyncioConnection

class _AmqpAsyncConsumer:
    ""This is an example consumer that will handle unexpected interactions
    with RabbitMQ such as channel and connection closures.

    If RabbitMQ closes the connection, this class will stop and indicate
    that reconnection is necessary. You should look at the output, as
    there are limited reasons why the connection may be closed, which
    usually are tied to permission related issues or socket timeouts.

    If the channel is closed, it will indicate a problem with one of the
    commands that were issued and that should surface in the output as well.

    ""

    EXCHANGE = "red_warden_rpc"

    def __init__(self):
        ""Create a new instance of the consumer class, passing in the AMQP
        URL used to connect to RabbitMQ.
        ""
        self.was_consuming = False

        self._connection = None
        self._channel = None
        self._closing = False
        self._close_event = asyncio.Event()

        # In production, experiment with higher prefetch values
        # for higher consumer throughput
        self._prefetch_count = 1
        self.queue_routes = []

    async def run(self):
        LOGGER.info("Connecting to RabbitMQ...")
        self._connection = AsyncioConnection(
            parameters=pika.URLParameters(
                url=get_random_element(Config["RW_AMQP_BROKER_HOST"])
            ),
            on_open_callback=self.on_connection_open,
            on_open_error_callback=self.on_connection_open_error,
            on_close_callback=self.on_connection_closed,
        )

        await self._close_event.wait()

    def on_connection_open(self, _unused_connection):
        LOGGER.info("Connection opened")
        self.open_channel()

    def on_connection_open_error(self, _unused_connection, err):
        LOGGER.error("Connection open failed: %s", err)
        self._close_event.set()

    def on_connection_closed(self, _unused_connection, reason):
        if not self._closing:
            LOGGER.warning("Connection closed, reconnect necessary: %s", reason)
            self._close_event.set()

    def close(self):
        if not self._closing:
            self._closing = True
            if self._connection:
                if self._connection.is_closing or self._connection.is_closed:
                    LOGGER.info("Connection is closing or already closed")
                else:
                    LOGGER.info("Closing connection")
                    self._connection.close()
            self._close_event.set()

    def open_channel(self):
        LOGGER.info("Creating a new channel")
        self._connection.channel(on_open_callback=self.on_channel_open)

    def on_channel_open(self, channel):
        LOGGER.info("Channel opened")
        self._channel = channel
        self.add_on_channel_close_callback()
        self.setup_exchange()

    def add_on_channel_close_callback(self):
        LOGGER.info("Adding channel close callback")
        self._channel.add_on_close_callback(self.on_channel_closed)

    def on_channel_closed(self, channel, reason):
        LOGGER.info("Channel %i was closed: %s", channel, reason)
        self.close()

    def setup_exchange(self):
        LOGGER.info("Declaring exchange: %s", self.EXCHANGE)
        self._channel.exchange_declare(
            exchange=self.EXCHANGE,
            exchange_type="topic",
            durable=True,
            callback=self.on_exchange_declareok,
        )

    def on_exchange_declareok(self, _unused_frame):
        LOGGER.info("Exchange declared: %s", self.EXCHANGE)
        self.setup_queues()

    def setup_queues(self):
        for queue_route in self.queue_routes:
            LOGGER.info("Declaring queue %s", queue_route["name"])
            cb = functools.partial(
                self.on_queue_declareok,
                userdata=(queue_route["name"], "RW.%s.*" % queue_route["topic"]),
            )
            self._channel.queue_declare(
                queue=queue_route["name"], durable=True, callback=cb
            )

        self.set_qos()

    def on_queue_declareok(self, _unused_frame, userdata):
        queue_name, routing_key = userdata

        LOGGER.info("Binding %s to %s with %s", self.EXCHANGE, queue_name, routing_key)
        self._channel.queue_bind(
            queue_name,
            self.EXCHANGE,
            routing_key=routing_key,
        )

    def set_qos(self):
        self._channel.basic_qos(
            prefetch_count=self._prefetch_count, callback=self.on_basic_qos_ok
        )

    def on_basic_qos_ok(self, _unused_frame):
        LOGGER.info("QOS set to: %d", self._prefetch_count)
        self.start_consuming()

    def start_consuming(self):
        LOGGER.info("Issuing consumer related RPC commands")
        self.add_on_cancel_callback()

        for queue_route in self.queue_routes:
            self._channel.basic_consume(queue_route["name"], self.on_message)

        self.was_consuming = True

    def on_message(self, _unused_channel, basic_deliver, properties, body):
        LOGGER.info(
            "Acknowledging message %s %s",
            basic_deliver.delivery_tag,
            basic_deliver.routing_key,
        )

        # TODO do this AFTER processing the message
        self._channel.basic_ack(basic_deliver.delivery_tag)

        try:
            # 1. find route by name >> basic_deliver.routing_key
            rpc = next(
                (
                    r
                    for name, r in Router.RPCs.items()
                    if name == basic_deliver.routing_key
                ),
                None,
            )
            if not rpc:
                raise RWNotFoundException

            req = json.loads(body.decode()) if body else {}

            ""----
            (
                _,
                controller_name,
                action_name,
            ) = basic_deliver.routing_key.rsplit(".")

            route = next(
                (
                    r
                    for r in Routes
                    if r["controller_name"] == controller_name
                    and r["action_name"] == action_name
                ),
                None,
            )
            if not route:
                raise RWNotFoundException(basic_deliver.routing_key)
            ""------------

            options = req.get("options", {}) or {}
            if not "internal" in options:
                options["internal"] = True

            rpc_info = {
                "route": route,
                "params": req.get("params", {}) or {},
                "query": req.get("query", {}) or {},
                "data": req.get("data", {}) or {},
                "identity": req.get("identity", {}) or {},
                "options": options,
            }

            results = self._rpc_manager.execution_callback(rpc_info)

        except Exception as err:
            results = self._rpc_manager.error_callback(err)

        self._channel.basic_publish(
            "",  # no exchange, the queue we are sending to is exclusive
            routing_key=properties.reply_to,
            properties=pika.BasicProperties(correlation_id=properties.correlation_id),
            body=json.dumps(results, cls=CustomJsonEncoder),
        )
        LOGGER.info(
            "Received message # %s from %s, reply to %s",
            basic_deliver.delivery_tag,
            properties.app_id,
            properties.reply_to,
        )
        LOGGER.debug(
            "Body of message # %s %s",
            basic_deliver.delivery_tag,
            body,
        )

    def add_on_cancel_callback(self):
        LOGGER.info("Adding consumer cancellation callback")
        self._channel.add_on_cancel_callback(self.on_consumer_cancelled)

    def on_consumer_cancelled(self, method_frame):
        LOGGER.info("Consumer was cancelled remotely, shutting down: %r", method_frame)
        if self._channel:
            self._channel.close()


class AmqpEngine(_RWEngine):
    def __init__(self):
        self._reconnect_delay = 0
        self._ac = _AmqpAsyncConsumer()

    def prepare_routes(self):
        for zone_controller in Router.zone_controllers:
            queue_route = {
                "name": "%s_%s_%s"
                % (
                    Config["RW_APP_NAME"],
                    Config["RW_MODULE_NAME"],
                    zone_controller,
                ),
                "topic": zone_controller,
            }
            self._ac.queue_routes.append(queue_route)

    def _check_reconnect_delay(self):
        if self._ac.was_consuming:
            self._reconnect_delay = 0
        else:
            self._reconnect_delay += 1

        if self._reconnect_delay > 30:
            self._reconnect_delay = 30

    async def run(self):
        while not self._shutdown_event.is_set():
            await self._ac.run()

            if not self._shutdown_event.is_set():
                self._check_reconnect_delay()
                LOGGER.info("Reconnecting after %d seconds", self._reconnect_delay)
                await asyncio.sleep(self._reconnect_delay)
                self._ac = _AmqpAsyncConsumer()

    def stop(self):
        self._shutdown_event.set()
        self._ac.close()

"""



from red_warden.graphql import (
    RWGraphQLConnection,
    RWGraphQLStringFilter,
)
from models import DatapathModel
from schemas.datapath.types import Datapath


class DatapathConnection(RWGraphQLConnection):
    class Meta:
        node = Datapath
        filters = {"default": {"name": RWGraphQLStringFilter()}}

    @staticmethod
    def resolve_all(_, info, **args):
        return DatapathModel.select()

    @staticmethod
    async def resolve_all_by_backend(parent, info, **args):
        return DatapathModel.select().where(DatapathModel.c.backend_id == parent.id)



from enum import Enum
import json
from functools import wraps
import datetime
from red_warden.helpers import import_string
import random
from cryptography.fernet import Fernet, InvalidToken
from red_warden.config import RWConfig
from red_warden.helpers import RWCustomJsonEncoder
from red_warden.datalayer import RWDataLayer, RWRedisBackend, RWMysqlBackend


RWConfig.env_string("RW_SESSIONS_TYPE", "cookie")
RWConfig.env_string("RW_SESSIONS_BACKEND_NAME")
RWConfig.env_string("RW_SESSIONS_NAME", "rws")
RWConfig.env_string("RW_SESSIONS_DOMAIN")
RWConfig.env_string("RW_SESSIONS_SAME_SITE", "lax")
RWConfig.env_bool("RW_SESSIONS_HTTPS_ONLY", True)
RWConfig.env_secret("RW_SESSIONS_SECRET")
RWConfig.env_int(
    "RW_SESSIONS_TIMEOUT", 14 * 24 * 60 * 60
)  # two weeks default expiration

if not RWConfig["RW_SESSIONS_SECRET"]:
    raise Exception(
        """
Missing SESSION secret key (env: RW_SESSIONS_SECRET). Try with one of these:
%s
%s
%s
"""
        % (
            str(Fernet.generate_key()),
            str(Fernet.generate_key()),
            str(Fernet.generate_key()),
        )
    )


class RWSessionTypeEnum(str, Enum):
    REDIS = "redis"
    MYSQL = "mysql"
    COOKIE = "cookie"


class RWMysqlSessionBackend:
    def __init__(self):
        if not RWConfig["RW_SESSIONS_BACKEND_NAME"]:
            raise KeyError("Mysql session backend not configured")

        backend = RWDataLayer._backends.get(RWConfig["RW_SESSIONS_BACKEND_NAME"], None)
        if not backend:
            raise KeyError(
                "Mysql session backend %s doesn't exists"
                % RWConfig["RW_SESSIONS_BACKEND_NAME"]
            )

        self.model = import_string("red_warden.datazones.RWManager.RWSessionModel")
        self.mysql = RWMysqlBackend(backend["endpoints"], backend["params"])

    async def connect(self):
        await self.mysql.connect()

    async def disconnect(self):
        await self.mysql.disconnect()

    async def get(self, key):

        row = await self.mysql.fetch_one(
            self.model.select().where(self.model.c.key == key)
        )
        if not row:
            return None
        else:
            row = await self.model.after_load(self.mysql, row)
            return row["value"]

    async def set(self, key, value, expires_at):
        # 30% chance to rotate key key, to avoid session fixation
        rotate_key = True if random.choice([0, 10]) < 3 else False
        if rotate_key:
            # let's create another key, delete the old one if exists
            if key:
                await self.delete(key)

            key = None

        if not key:
            key = Fernet.generate_key().decode()

        data = {
            "key": key,
            "value": value,
            "expires_at": datetime.datetime.now() + datetime.timedelta(0, expires_at),
        }

        should_create = True
        if key:
            updated_row = await self.mysql.execute(
                self.model.update()
                .where(self.model.c.key == key)
                .values(await self.model.before_update(self.mysql, data))
            )
            if updated_row > 0:
                should_create = False

        if should_create:
            data = await self.model.before_create(self.mysql, data)
            await self.mysql.execute(self.model.insert().values(data))

        return key

    async def delete(self, key):
        await self.mysql.execute(self.model.delete().where(self.model.c.key == key))


class RWRedisSessionBackend:
    def __init__(self):
        if not RWConfig["RW_SESSIONS_BACKEND_NAME"]:
            raise KeyError("Redis session backend not configured")

        backend = RWDataLayer._backends.get(RWConfig["RW_SESSIONS_BACKEND_NAME"], None)
        if not backend:
            raise KeyError(
                "Redis session backend %s doesn't exists"
                % RWConfig["RW_SESSIONS_BACKEND_NAME"]
            )

        self.redis = RWRedisBackend(backend["endpoints"], backend["params"])
        self.fernet = Fernet(str(RWConfig["RW_SESSIONS_SECRET"]))

    async def connect(self):
        await self.redis.connect()

    async def disconnect(self):
        await self.redis.disconnect()

    async def get(self, key):
        encrypted = await self.redis.get(key)
        if not encrypted:
            return None
        return json.loads(self.fernet.decrypt(encrypted.encode("utf-8")))

    async def set(self, key, value, expires_at):
        rotate_key = True if random.choice([0, 10]) < 3 else False
        if rotate_key:
            # let's create another key, delete the old one if exists
            if key:
                await self.delete(key)
            key = None

        if not key:
            key = "sessions:%s" % Fernet.generate_key().decode()

        s = json.dumps(value, cls=RWCustomJsonEncoder)
        await self.redis.set(key, self.fernet.encrypt(s.encode()), expire=expires_at)
        return key

    async def delete(self, key):
        await self.redis.delete(key)


async def _wrap_with_session(func, request):
    if RWConfig["RW_SESSIONS_TYPE"] == RWSessionTypeEnum.REDIS:
        backend = RWRedisSessionBackend()
    elif RWConfig["RW_SESSIONS_TYPE"] == RWSessionTypeEnum.MYSQL:
        backend = RWMysqlSessionBackend()
    else:
        backend = None

    if backend:
        await backend.connect()

    max_age = RWConfig["RW_SESSIONS_TIMEOUT"]
    fernet = Fernet(str(RWConfig["RW_SESSIONS_SECRET"]))
    initial_session_was_empty = True
    cookie = request.cookies.get(RWConfig["RW_SESSIONS_NAME"])
    if cookie:
        try:
            data = fernet.decrypt(cookie.encode("utf-8"), max_age).decode("utf-8")

            if RWConfig["RW_SESSIONS_TYPE"] == RWSessionTypeEnum.COOKIE:
                request.state.session = json.loads(data)
                initial_session_was_empty = False
            else:
                session_key = json.loads(data).get("sid")
                request.state.session = await backend.get(session_key.encode("utf-8"))
                if not request.state.session:
                    request.state.session = {}
                    initial_session_was_empty = True

                request.state.session_key = session_key

            initial_session_was_empty = False
        except InvalidToken:
            request.state.session = {}
    else:
        request.state.session = {}

    response = await func(request)

    session_key = getattr(request.state, "session_key", None)
    if request.state.session:
        if RWConfig["RW_SESSIONS_TYPE"] == RWSessionTypeEnum.COOKIE:
            cookie_data = request.state.session
        else:
            session_key = await backend.set(session_key, request.state.session, max_age)
            cookie_data = {"sid": session_key}

        data = fernet.encrypt(json.dumps(cookie_data).encode("utf-8")).decode("utf-8")

        response.set_cookie(
            key=RWConfig["RW_SESSIONS_NAME"],
            value=data,
            secure=RWConfig["RW_SESSIONS_HTTPS_ONLY"],
            max_age=max_age,
            samesite=RWConfig["RW_SESSIONS_SAME_SITE"],
            httponly=True,
            domain=RWConfig["RW_SESSIONS_DOMAIN"],
        )

    elif not initial_session_was_empty:
        if backend:
            await backend.delete(session_key)

        response.set_cookie(
            key=RWConfig["RW_SESSIONS_NAME"],
            value="",
            secure=RWConfig["RW_SESSIONS_HTTPS_ONLY"],
            expires=0,
            samesite=RWConfig["RW_SESSIONS_SAME_SITE"],
            httponly=True,
            domain=RWConfig["RW_SESSIONS_DOMAIN"],
        )

    if backend:
        await backend.disconnect()

    return response


def session_from_cookie(_func=None):
    def decorator_use_session(func):
        @wraps(func)
        async def wrapper_use_session(request):
            return await _wrap_with_session(func, request)

        return wrapper_use_session

    if _func is None:
        return decorator_use_session
    else:
        return decorator_use_session(_func)



from abc import ABC
import os
import base64
import aioredis
import inspect
import importlib.util
import databases
import random
import json
import copy
from cryptography.fernet import Fernet
import sqlalchemy
from red_warden.config import RWConfig
from red_warden.helpers import get_random_element
from red_warden.enums import (
    RWMultiTenancyEnum,
    RWBackendTypeEnum,
    RWBackendDestinationEnum,
)
from red_warden.helpers import import_string


class RWDataLayerException(Exception):
    pass


RWConfig.env_secret("RW_DATABASE_SECRET")
RWConfig.env_bool("RW_DATABASE_DEBUG", False)
if not RWConfig["RW_DATABASE_SECRET"]:
    raise RWDataLayerException(
        """
Missing DATABASE secret key (env: RW_DATABASE_SECRET). Try with one of these:
%s
%s
%s
        """
        % (
            str(Fernet.generate_key()),
            str(Fernet.generate_key()),
            str(Fernet.generate_key()),
        )
    )


# region --- Backend Classes ---
class RWMysqlBackend(databases.Database):
    def __init__(self, endpoints, params):
        if isinstance(endpoints, str):
            self.endpoints = endpoints.split(",")
        elif not isinstance(endpoints, list):
            self.endpoints = [endpoints]
        else:
            self.endpoints = endpoints

        self.params = params

        username = self.params.get("username", None)
        password = self.params.get("password", None)
        url = "mysql://"
        if username or password:
            url += "%s:%s" % (username, password)

        url += "@%s" % get_random_element(self.endpoints)

        db_name = self.params.get("db_name", None)
        if db_name:
            url += "/%s" % db_name

        self.server_code = self.params.get("server_code", "0")
        super().__init__(url, echo=RWConfig["RW_DATABASE_DEBUG"])

    async def get_uuid(self):
        uuid = await self.fetch_val("SELECT UUID() as uuid")
        if self.server_code is not None:
            uuid = uuid[0:24] + str(self.server_code).zfill(4) + uuid[28:]

        uuid_rev = uuid[24:] + uuid[19:23] + uuid[14:18] + uuid[9:13] + uuid[:8]
        return uuid_rev.upper()


class RWRedisBackend:
    _redis = None

    def __init__(self, endpoints, params):
        if isinstance(endpoints, str):
            endpoints = endpoints.split(",")
        elif not isinstance(endpoints, list):
            endpoints = [endpoints]

        self._url = "redis://%s/%s" % (
            get_random_element(endpoints),
            params.get("db", 0),
        )

    async def __aenter__(self):
        await self.connect()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.disconnect()

    async def connect(self):
        self._redis = await aioredis.create_redis(self._url, encoding="utf-8")

    async def disconnect(self):
        if self._redis:
            self._redis.close()
            await self._redis.wait_closed()
            self._redis = None

    @property
    def is_connected(self):
        return True if self._redis else False

    def __getattr__(self, item):
        if not self._redis:
            raise Exception("Redis is not connected")

        return getattr(self._redis, item)


# endregion

# region --- Datazone classes ---
class RWDatazone(ABC):
    models = []
    controllers = []
    multi_tenancy = RWMultiTenancyEnum.GLOBAL
    schema_path = None

    def __init_subclass__(cls, **kwargs):
        cls.schema_path = os.path.join(
            os.path.dirname(os.path.realpath(inspect.getmodule(cls).__file__)),
            "migrations",
        )


# endregion


class RWDataLayer:
    _backends = {}
    _datazones = {}
    _models = {}

    # region --- Backend Methods ---
    @classmethod
    def add_backend(
        cls,
        backend_type,
        backend_id,
        backend_name,
        backend_destination,
        endpoints,
        params,
    ):
        if backend_name in cls._backends.keys():
            raise RWDataLayerException("Backend %s already exists" % backend_name)

        if not endpoints:
            raise RWDataLayerException(
                "No endpoints specified for backend %s" % backend_name
            )

        if isinstance(endpoints, str):
            endpoints = endpoints.split(",")
        elif not isinstance(endpoints, list):
            endpoints = [endpoints]

        if backend_type == RWBackendTypeEnum.MYSQL:
            backend_class = RWMysqlBackend
        elif backend_type == RWBackendTypeEnum.REDIS:
            backend_class = RWRedisBackend
        else:
            raise NotImplementedError

        cls._backends[backend_name] = {
            "id": backend_id,
            "name": backend_name,
            "type": backend_type,
            "destination": backend_destination,
            "class": backend_class,
            "endpoints": endpoints,
            "params": params or {},
        }

    # endregion

    # region --- Datazone Methods ---
    @classmethod
    def get_datazones(cls):
        return cls._datazones

    @classmethod
    def add_datazone(
        cls,
        datazone_name,
        models,
        controllers,
        multi_tenancy,
        schema_path=None,
    ):
        if datazone_name not in cls._datazones.keys():
            cls._datazones[datazone_name] = {
                "multi_tenancy": multi_tenancy,
                "schema_path": schema_path,
                "controllers": controllers,
            }

        for model in models:
            cls._models[model.__name__] = datazone_name

    @classmethod
    def get_datazone_from_model(cls, model):
        datazone = cls._models.get(model.__name__, None)
        if not datazone:
            raise Exception("No datazone for model %s" % model.__name__)
        return datazone

    # @classmethod
    # def get_datazone_tenant_params(cls, datazone_name, tenant_id):
    #     if datazone_name not in cls._datazones.keys():
    #         raise Exception("Datazone %s does not exists." % datazone_name)
    #
    #     if tenant_id not in cls._datazones[datazone_name]["tenants"]:
    #         raise Exception(
    #             "Configuration %s for datazone %s does not exists."
    #             % (tenant_id, datazone_name)
    #         )
    #
    #     return cls._datazones[datazone_name]["tenants"][tenant_id]
    #
    # @classmethod
    # def set_datazone_tenant_params(cls, datazone_name, tenant_id, tenant_params):
    #     tenant_id = tenant_id or "GLOBAL"
    #     if datazone_name not in cls._datazones.keys():
    #         raise Exception("Datazone %s does not exists." % datazone_name)
    #
    #     cls._datazones[datazone_name][tenant_id] = tenant_params
    # endregion

    @classmethod
    async def _get_datapath(cls, datazone_name, tenant_id):
        """
        # TODO write this better: datazone and tenant_id are UNIQUE as a pair
        """
        if isinstance(tenant_id, str):
            tenant_id = bytes.fromhex(tenant_id)

        params = {
            "username": RWConfig["RW_BACKEND_USERNAME"],
            "password": RWConfig["RW_BACKEND_PASSWORD"],
            "db_name": RWConfig["RW_BACKEND_DB_NAME"],
        }
        params.update(json.loads(RWConfig["RW_BACKEND_PARAMS"]))
        async with RWMysqlBackend(RWConfig["RW_BACKEND_ENDPOINTS"], params) as db:
            model = import_string("red_warden.datazones.RWManager.RWDatapathModel")
            datapath = await db.fetch_one(
                model.select().where(
                    sqlalchemy.and_(
                        model.c.datazone == datazone_name,
                        model.c.tenant_id == tenant_id,
                    )
                )
            )

            if not datapath:
                return None
            else:
                datapath = await model.after_load(db, datapath)
                datapath["backend"] = next(
                    (
                        backend
                        for backend_name, backend in cls._backends.items()
                        if backend["id"] == datapath["backend_id"]
                    ),
                    None,
                )

                if datapath["backend"] and tenant_id:
                    model = import_string(
                        "red_warden.datazones.RWManager.RWTenantModel"
                    )
                    tenant = await db.fetch_one(
                        model.select().where(
                            model.c.id == bytes.fromhex(datapath["tenant_id"])
                        )
                    )
                    if tenant:
                        tenant = await model.after_load(db, tenant)
                        datapath["backend"]["params"].update(tenant["params"])

                return datapath

    @classmethod
    async def _set_datapath(cls, datazone, tenant_id, backend_name, version):
        params = {
            "username": RWConfig["RW_BACKEND_USERNAME"],
            "password": RWConfig["RW_BACKEND_PASSWORD"],
            "db_name": RWConfig["RW_BACKEND_DB_NAME"],
        }
        params.update(json.loads(RWConfig["RW_BACKEND_PARAMS"]))
        async with RWMysqlBackend(RWConfig["RW_BACKEND_ENDPOINTS"], params) as db:
            model = import_string("red_warden.datazones.RWManager.RWBackendModel")
            backend = await db.fetch_one(
                model.select().where(model.c.name == backend_name)
            )

            model = import_string("red_warden.datazones.RWManager.RWDatapathModel")
            datapath = await db.fetch_one(
                model.select().where(
                    sqlalchemy.and_(
                        model.c.datazone == datazone, model.c.tenant_id == tenant_id
                    )
                )
            )

            data = {
                "datazone": datazone,
                "backend_id": bytes.hex(backend["id"]),
                "tenant_id": tenant_id,
                "version": version,
            }
            if not datapath:
                data = await model.before_create(db, data, True)
                await db.execute(model.insert().values(data))
            else:
                data = await model.before_update(db, data, True)
                await db.execute(
                    model.update().where(model.c.id == datapath["id"]).values(data)
                )

    @classmethod
    async def _execute_datapath_schema_migrations(cls, datazone, actual_version, db):
        version = actual_version
        datazone_path = cls._datazones[datazone]["schema_path"]
        if os.path.exists(datazone_path):
            files = os.listdir(datazone_path)
            files.sort()

            for file in files:
                name, ext = os.path.splitext(file)
                if ext == ".py" and int(name) > version:
                    version = int(name)
                    spec = importlib.util.spec_from_file_location(
                        "migration", os.path.join(datazone_path, file)
                    )
                    migration = importlib.util.module_from_spec(spec)
                    spec.loader.exec_module(migration)
                    await migration.upgrade(db)

        return version

    @classmethod
    async def upgrade_manager_datapath_schema(cls):
        endpoints = RWConfig["RW_BACKEND_ENDPOINTS"]
        params = {
            "username": RWConfig["RW_BACKEND_USERNAME"],
            "password": RWConfig["RW_BACKEND_PASSWORD"],
            "db_name": RWConfig["RW_BACKEND_DB_NAME"],
        }
        params.update(json.loads(RWConfig["RW_BACKEND_PARAMS"]))

        datapath = None
        db = RWMysqlBackend(endpoints, params)
        try:
            async with db:
                datapath_table_exists = await db.fetch_all(
                    "SHOW TABLES LIKE 'datapath'"
                )
                if datapath_table_exists:
                    datapath = await cls._get_datapath("RWManager", None)
        except:
            pass

        new_version = await cls._execute_datapath_schema_migrations(
            "RWManager", datapath["version"] if datapath else 0, db
        )

        if not datapath or (datapath["version"] != new_version and new_version > 0):
            await cls._set_datapath("RWManager", None, "rw_global", new_version)

        await cls.reload_backends()

    @classmethod
    async def _create_mysql_db_and_upgrade_datapath_migrations(
        cls, datazone_name, tenant_id
    ):
        datapath = await cls._get_datapath(datazone_name, tenant_id)

        if datapath:
            backend = datapath["backend"]
            if tenant_id:
                backend["params"]["db_name"] = tenant_id

        elif tenant_id:
            # we have a tenant_id, but no backend configured.
            # Let's choose a random one from the TENANT_AVAILABLE backends
            tenant_backends = list(
                backend
                for backend_name, backend in cls._backends.items()
                if backend["type"] == RWBackendTypeEnum.MYSQL
                and backend["destination"] == RWBackendDestinationEnum.TENANTS_AVAILABLE
            )
            if not tenant_backends:
                raise RWDataLayerException("No backend available for tenant creation")

            backend = random.choice(tenant_backends)
            backend["params"]["db_name"] = tenant_id
        else:
            # we have no datapath and no tenant, so we need a global backend.
            backend = cls._backends["rw_global"]

        # creates the designated database, if not present
        async with RWMysqlBackend(
            backend["endpoints"],
            {
                "username": backend["params"].get("username", None),
                "password": backend["params"].get("password", None),
            },
        ) as db:
            # check if database exists, else creates it
            db_exists = await db.fetch_val(
                "SHOW DATABASES LIKE :db_name",
                {"db_name": backend["params"].get("db_name", None)},
            )
            if not db_exists:
                await db.execute(
                    "CREATE DATABASE %s DEFAULT CHARACTER SET utf8mb4 COLLATE utf8mb4_0900_ai_ci"
                    % backend["params"].get("db_name", None)
                )

                # ensure all grants are set for this user to the Manager database and tables
                await db.execute(
                    "GRANT ALL ON %s.* TO :username@'%%'"
                    % backend["params"].get("db_name", None),
                    {"username": backend["params"].get("username", None)},
                )

        old_version = datapath["version"] if datapath else 0
        async with RWMysqlBackend(backend["endpoints"], backend["params"]) as db:
            new_version = await cls._execute_datapath_schema_migrations(
                datazone_name, old_version, db
            )
        if not datapath or (datapath["version"] != new_version and new_version > 0):
            await cls._set_datapath(
                datazone_name, tenant_id, backend["name"], new_version
            )

        if old_version != new_version:
            return {
                "datazone": datazone_name,
                "tenant": tenant_id,
                "old_version": datapath["version"] if datapath else 0,
                "new_version": new_version,
            }
        else:
            return None

    @classmethod
    async def upgrade_other_datapaths_migrations(cls, tenant_id=None):
        results = []

        for datazone_name, datazone in cls._datazones.items():
            if datazone_name != "RWManager":
                if datazone["multi_tenancy"] == RWMultiTenancyEnum.GLOBAL:
                    result = await cls._create_mysql_db_and_upgrade_datapath_migrations(
                        datazone_name, None
                    )
                    if result:
                        results.append(result)

                else:  # multi_tenancy = RWMultiTenancyEnum.TENANT
                    params = {
                        "username": RWConfig["RW_BACKEND_USERNAME"],
                        "password": RWConfig["RW_BACKEND_PASSWORD"],
                        "db_name": RWConfig["RW_BACKEND_DB_NAME"],
                    }
                    params.update(json.loads(RWConfig["RW_BACKEND_PARAMS"]))
                    async with RWMysqlBackend(
                        RWConfig["RW_BACKEND_ENDPOINTS"], params
                    ) as db:
                        if tenant_id:
                            result = await cls._create_mysql_db_and_upgrade_datapath_migrations(
                                datazone_name, tenant_id
                            )
                            if result:
                                results.append(result)
                        else:
                            model = import_string(
                                "red_warden.datazones.RWManager.RWTenantModel"
                            )
                            tenants = await db.fetch_all(model.select())
                            for t in tenants:
                                t = await model.after_load(db, t)
                                result = await cls._create_mysql_db_and_upgrade_datapath_migrations(
                                    datazone_name, t["id"]
                                )
                                if result:
                                    results.append(result)

        return results

    @classmethod
    async def reload_backends(cls):
        cls._backends = {}

        # check for datapath table, it manages the schema versioning
        params = {
            "username": RWConfig["RW_BACKEND_USERNAME"],
            "password": RWConfig["RW_BACKEND_PASSWORD"],
            "db_name": RWConfig["RW_BACKEND_DB_NAME"],
        }
        params.update(json.loads(RWConfig["RW_BACKEND_PARAMS"]))
        async with RWMysqlBackend(RWConfig["RW_BACKEND_ENDPOINTS"], params) as db:
            model = import_string("red_warden.datazones.RWManager.RWBackendModel")
            backends = await db.fetch_all(model.select().where(model.c.enabled == "Y"))
            for b in backends:
                b = await model.after_load(db, b)
                if b["type"] == "mysql":
                    backend_type = RWBackendTypeEnum.MYSQL
                # elif b.type == "mongodb":  TODO add mongodb support
                #    backend_class = RWMongoDbBackend
                elif b["type"] == "redis":
                    backend_type = RWBackendTypeEnum.REDIS

                cls.add_backend(
                    backend_type,
                    b["id"],
                    b["name"],
                    b["destination"],
                    b["endpoints"],
                    b["params"],
                )

    @classmethod
    async def get_backend_from_datapath(cls, datazone_name, tenant_id=None):
        if datazone_name not in cls._datazones.keys():
            raise RWDataLayerException("Datazone %s does not exist" % datazone_name)

        datapath = await cls._get_datapath(datazone_name, tenant_id)
        new_backend = copy.deepcopy(datapath["backend"])
        return new_backend

        # try:
        #     tenant_params = cls.get_datazone_tenant_params(datazone_name, tenant_id)
        # except:
        #     pass
        #
        # backend = cls._backends.get(backend_name, None)
        # if not backend:
        #     raise RWBackendException("Backend %s does not exist" % backend_name)
        #
        # endpoint = endpoint or choice(backend["endpoints"])
        # params = backend["params"]
        # if additional_params:
        #     params.update(additional_params)
        #
        # if not hasattr(request.state, "backends"):
        #     request.state.backends = {}
        #
        # instance_hash = cls._calc_instance_hash(endpoint, params)
        # if instance_hash not in request.state.backends.keys():
        #     request.state.backends[instance_hash] = backend["class"](endpoint, params)
        #
        # return request.state.backends[instance_hash]

    @staticmethod
    def get_sort_columns(table, sort_args):
        if isinstance(sort_args, str):
            sort_args = [sort_args]

        sort = []
        for s in sort_args:
            if s:
                if s[0] == "-":
                    s = s[1:]
                    desc = True
                else:
                    desc = False

                if not hasattr(table.c, s):
                    raise Exception("Wrong sort field: %s" % s)
                if desc:
                    sort.append(table.c[s].desc())
                else:
                    sort.append(table.c[s])

        if not sort:
            sort = [table.primary_key[0]]

        return sort

    @staticmethod
    def get_filters(model, table, filters_args):
        filters = []
        special_columns = model.get_special_columns()
        for f in filters_args:
            k_op_v = f.split(":")
            if len(k_op_v) == 2:
                k = k_op_v[0]
                op = "eq"
                v = k_op_v[1]
            elif len(k_op_v) == 3:
                k, op, v = k_op_v
            else:
                raise RWDataLayerException("Wrong filter definition: %s" % f)

            if not hasattr(table.c, k):
                raise RWDataLayerException("Wrong filter field: %s" % k)

            if v == "None":
                v = None
            elif k in special_columns["id"]:
                v = bytes.fromhex(v)

            if op == "eq":
                filters.append(table.c[k] == v)
            elif op == "ne":
                filters.append(table.c[k] != v)
            elif op == "lt":
                filters.append(table.c[k] < v)
            elif op == "lte":
                filters.append(table.c[k] <= v)
            elif op == "gt":
                filters.append(table.c[k] > v)
            elif op == "gte":
                filters.append(table.c[k] >= v)
            elif op == "like":
                filters.append(table.c[k].like("%" + v + "%"))
            elif op == "ilike":
                filters.append(table.c[k].ilike("%" + v + "%"))
            elif op == "not_like":
                filters.append(~table.c[k].like("%" + v + "%"))
            elif op == "not_ilike":
                filters.append(~table.c[k].ilike("%" + v + "%"))
            elif op == "in":
                filters.append(table.c[k].in_(v.split("|")))
            elif op == "not_in":
                filters.append(~table.c[k].in_(v.split("|")))
            else:
                raise RWDataLayerException("Wrong filter operator: %s" % op)

        return filters

    @classmethod
    async def load_by_id(cls, model, model_id, identity=None):
        tenant_id = identity.get("tenant_id", None) if identity else None

        if isinstance(model_id, str):
            try:
                int(model_id, 16)
                model_id = bytes.fromhex(model_id)
            except ValueError:
                raise RWDataLayerException("Malformed ID")

        datazone = cls.get_datazone_from_model(model)
        backend = await cls.get_backend_from_datapath(datazone, tenant_id)
        async with backend["class"](backend["endpoints"], backend["params"]) as db:
            query = model.get_select_query()

            row = await db.fetch_one(query.where(model.c.id == model_id))
            if row:
                return await model.after_load(db, row)
            else:
                return None

    @classmethod
    async def load_one(cls, model, identity=None, filters=None):
        filters = filters or []
        tenant_id = identity.get("tenant_id", None) if identity else None
        if isinstance(model, str):
            model = import_string(model)

        datazone = cls.get_datazone_from_model(model)
        backend = await cls.get_backend_from_datapath(datazone, tenant_id)
        async with backend["class"](backend["endpoints"], backend["params"]) as db:
            query = model.get_select_query().cte("resolved")
            filter_columns = cls.get_filters(model, query, filters)
            cte = (
                sqlalchemy.select(
                    [
                        *query.c,
                    ]
                )
                .where(sqlalchemy.and_(*filter_columns))
                .cte("ordered")
            )
            row = await db.fetch_one(cte.select())
            if row:
                return await model.after_load(db, row)
            else:
                return None

    @classmethod
    async def load_all(
        cls, model, identity=None, pagination=None, sort=None, filters=None
    ):
        sort = sort or []
        filters = filters or []
        pagination = pagination or {"after": None, "before": None, "limit": 25}
        tenant_id = identity.get("tenant_id", None) if identity else None

        if isinstance(model, str):
            model = import_string(model)

        datazone = cls.get_datazone_from_model(model)
        backend = await cls.get_backend_from_datapath(datazone, tenant_id)
        async with backend["class"](backend["endpoints"], backend["params"]) as db:
            query = model.get_select_query()

            cte_resolved = query.cte("resolved")

            try:
                limit = int(pagination["limit"]) or 25
                if limit > 500:
                    limit = 500
            except ValueError:
                limit = 25

            cursor_hint = pagination["after"] or pagination["before"]
            if cursor_hint:
                try:
                    cursor = json.loads(base64.b64decode(cursor_hint.encode()).decode())
                except Exception:
                    raise RWDataLayerException("Wrong cursor format")

                sort_columns = cls.get_sort_columns(cte_resolved, cursor["sort"])
                filter_columns = cls.get_filters(model, cte_resolved, cursor["filters"])
            else:
                cursor = None
                sort_columns = cls.get_sort_columns(cte_resolved, sort)
                filter_columns = cls.get_filters(model, cte_resolved, filters)

            cte_ordered = (
                sqlalchemy.select(
                    [
                        sqlalchemy.func.row_number()
                        .over(order_by=sort_columns)
                        .label("n"),
                        *cte_resolved.c,
                    ]
                )
                .where(sqlalchemy.and_(*filter_columns))
                .cte("ordered")
            )

            q = cte_ordered.select()
            total_count = await db.fetch_one(
                sqlalchemy.select([sqlalchemy.func.count().label("total")]).select_from(
                    q.alias("ordered")
                )
            )
            total_count = total_count["total"]

            if cursor:
                cte_id_check = (
                    sqlalchemy.select([cte_ordered.c.n]).where(
                        cte_ordered.c.id == bytes.fromhex(cursor["id"])
                    )
                ).cte("id_check")

                if pagination["after"]:
                    q = q.where(
                        cte_ordered.c.n
                        > sqlalchemy.sql.func.ifnull(cte_id_check.c.n, 0)
                    )
                elif pagination["before"]:
                    q = q.where(
                        cte_ordered.c.n
                        > sqlalchemy.sql.func.ifnull(cte_id_check.c.n, 0) - limit
                    )

            if limit:
                q = q.limit(limit)

            def encode_cursor(row):
                _cursor = {
                    "id": "".join("{:02x}".format(b) for b in row.id),
                }
                if cursor:
                    _cursor["filters"] = cursor.get("filters", {})
                    _cursor["sort"] = cursor.get("sort", {})
                else:
                    _cursor["filters"] = filters
                    _cursor["sort"] = sort
                return base64.b64encode(json.dumps(_cursor).encode()).decode()

            data = await db.fetch_all(q)

            if len(data):
                start_cursor = encode_cursor(data[0])
                end_cursor = encode_cursor(data[-1])
                has_previous_page = data[0].n > 1
                has_next_page = data[-1].n < total_count

            else:
                start_cursor = None
                end_cursor = None
                if pagination["after"]:
                    has_previous_page = total_count > 0
                    has_next_page = False
                elif pagination["before"]:
                    has_previous_page = False
                    has_next_page = total_count > 0
                else:
                    has_previous_page = False
                    has_next_page = False

            return [await model.after_load(db, row) for row in data], {
                "count": len(data),
                "total": total_count,
                "start": start_cursor,
                "end": end_cursor,
                "prev": has_previous_page,
                "next": has_next_page,
            }

    @classmethod
    async def save(cls, model, model_id, data, identity=None, skip_validation=False):
        tenant_id = identity.get("tenant_id", None) if identity else None

        if isinstance(model, str):
            model = import_string(model)

        datazone = cls.get_datazone_from_model(model)
        backend = await cls.get_backend_from_datapath(datazone, tenant_id)
        async with backend["class"](backend["endpoints"], backend["params"]) as db:
            if model_id:
                data = await model.before_update(db, data, skip_validation)
                await db.execute(
                    model.update()
                    .where(model.c.id == bytes.fromhex(model_id))
                    .values(data)
                )
            else:
                data = await model.before_create(db, data, skip_validation)
                await db.execute(model.insert().values(data))
                model_id = data["id"]

            return await cls.load_by_id(model, model_id, identity)

    @classmethod
    async def remove(cls, model, model_id, identity=None):
        tenant_id = identity.get("tenant_id", None) if identity else None

        if isinstance(model, str):
            model = import_string(model)

        datazone = cls.get_datazone_from_model(model)
        backend = await cls.get_backend_from_datapath(datazone, tenant_id)
        async with backend["class"](backend["endpoints"], backend["params"]) as db:
            row_count = await db.execute(
                model.delete().where(model.c.id == bytes.fromhex(model_id))
            )
            if row_count == 0:
                return None
            else:
                return model_id



from functools import wraps
from starlette.exceptions import HTTPException
from red_warden.helpers import import_string
from red_warden.datalayer import RWDataLayer


required_scope = {}


def _check_authorization(func, identity):
    func_name = "%s_%s" % (func.__self__.__name__, func.__name__)
    global required_scope
    if func_name in required_scope.keys():
        func_scopes = required_scope[func_name]
        if func_scopes:
            if not identity:
                raise HTTPException(401)

            user_scopes = identity.get("scope", [])

            # Check if any of the method scopes are verified.
            # the logic is:
            #
            # [ "scope_01", "scope_02" ]
            # needs BOTH scopes to be verified
            #
            # [ [ "scope_03" ], [ "scope_04", "scope_05" ]
            # is verified when user has "scope_03" OR both "scope_04" AND "scope_05"

            if isinstance(func_scopes, str):
                # single scope, make a list
                func_scopes = [func_scopes]

            oks = [True] * len(func_scopes)
            for i, func_scope in enumerate(func_scopes):
                if isinstance(func_scope, str):
                    func_scope = [func_scope]

                for fs in func_scope:
                    if fs not in user_scopes:
                        oks[i] = False
                        break

            if not any(oks):
                # at least one is True, no need to check them all
                raise HTTPException(403)


def auth_from_session(_func=None):
    def decorator_auth(func):
        @wraps(func)
        async def wrapper_auth(request):
            request.state.identity = request.state.session.get("identity", {})

            _check_authorization(func, request.state.identity)

            return await func(request)

        return wrapper_auth

    if _func is None:
        return decorator_auth
    else:
        return decorator_auth(_func)


def auth_from_bearer_token(_func=None):
    def decorator_auth(func):
        @wraps(func)
        async def wrapper_auth(*args):
            request = args[0]
            request.state.identity = {}

            auth_header = request.headers.get("authorization", None)
            if auth_header:
                try:
                    bearer, access_token = auth_header.split()
                    if bearer != "Bearer":
                        raise ValueError()

                    model = import_string(
                        "red_warden.datazones.RWManager.RWOAuth2TokenModel"
                    )
                    token = await RWDataLayer.load_one(
                        model, filters=["access_token:eq:%s" % access_token]
                    )
                    if token:
                        request.state.identity["key_id"] = token["key_id"]
                        request.state.identity["scope"] = token["scope"]

                except ValueError:
                    pass

            _check_authorization(func, request.state.identity)
            return await func(request)

        return wrapper_auth

    if _func is None:
        return decorator_auth
    else:
        return decorator_auth(_func)



import os
from cryptography.hazmat.primitives.kdf.scrypt import Scrypt
from cryptography.hazmat.backends import default_backend
from cryptography.exceptions import InvalidKey
from datetime import datetime
import pyotp
import sqlalchemy
from red_warden.config import RWConfig
from red_warden.mvc import RWMysqlModel, RWValidationModelBase
from red_warden.enums import (
    RWBackendTypeEnum,
    RWBackendDestinationEnum,
    RWYesNoEnum,
)
from typing import Optional


class RWBackendModel(RWMysqlModel):
    table_name = "backend"

    class ValidationModel(RWValidationModelBase):
        name: str
        type: RWBackendTypeEnum
        destination: RWBackendDestinationEnum
        params: Optional[str]
        endpoints: str
        enabled: RWYesNoEnum

    @staticmethod
    def get_db_columns():
        return [
            sqlalchemy.Column("name", sqlalchemy.String(length=250), nullable=False),
            sqlalchemy.Column(
                "type",
                sqlalchemy.Enum("mysql", "mongodb", "redis"),
                nullable=False,
            ),
            sqlalchemy.Column(
                "destination",
                sqlalchemy.Enum("global", "tenants_available", "tenants_full"),
                nullable=False,
            ),
            sqlalchemy.Column("params", sqlalchemy.String(length=250), nullable=False),
            sqlalchemy.Column(
                "endpoints", sqlalchemy.String(length=250), nullable=False
            ),
            sqlalchemy.Column(
                "enabled", sqlalchemy.Enum("Y", "N"), default="N", nullable=False
            ),
        ]

    @classmethod
    def get_special_columns(cls):
        sc = super().get_special_columns()
        sc["encrypted"].append("params")
        sc["json"].append("params")
        return sc


class RWSessionModel(RWMysqlModel):
    table_name = "session"

    class ValidationModel(RWValidationModelBase):
        key: str
        value: str
        expires_at: datetime

    @staticmethod
    def get_db_columns():
        return [
            sqlalchemy.Column("key", sqlalchemy.String(length=50), nullable=False),
            sqlalchemy.Column("value", sqlalchemy.String(length=2048), nullable=False),
            sqlalchemy.Column("expires_at", sqlalchemy.DateTime(), nullable=False),
        ]

    @staticmethod
    def get_secret():
        return str(RWConfig["RW_SESSIONS_SECRET"])

    @classmethod
    def get_special_columns(cls):
        sc = super().get_special_columns()
        sc["encrypted"].append("value")
        sc["json"].append("value")
        return sc


class RWTenantModel(RWMysqlModel):
    table_name = "tenant"

    class ValidationModel(RWValidationModelBase):
        name: str
        params: Optional[str]
        enabled: RWYesNoEnum

    @staticmethod
    def get_db_columns():
        return [
            sqlalchemy.Column("name", sqlalchemy.String(length=250), nullable=False),
            sqlalchemy.Column("params", sqlalchemy.String(length=250), nullable=False),
            sqlalchemy.Column(
                "enabled", sqlalchemy.Enum("Y", "N"), default="N", nullable=False
            ),
        ]

    @classmethod
    def get_special_columns(cls):
        sc = super().get_special_columns()
        sc["encrypted"].append("params")
        sc["json"].append("params")
        return sc


class RWDatapathModel(RWMysqlModel):
    table_name = "datapath"

    class ValidationModel(RWValidationModelBase):
        tenant_id: Optional[str]
        datazone: str
        backend_id: str
        db_name: Optional[str]
        version: int

    @staticmethod
    def get_db_columns():
        return [
            sqlalchemy.Column(
                "tenant_id",
                sqlalchemy.Binary(16),
                sqlalchemy.ForeignKey("tenant.id"),
                nullable=False,
            ),
            sqlalchemy.Column(
                "datazone", sqlalchemy.String(length=250), nullable=False
            ),
            sqlalchemy.Column(
                "backend_id",
                sqlalchemy.Binary(16),
                sqlalchemy.ForeignKey("backend.id"),
                nullable=False,
            ),
            sqlalchemy.Column(
                "db_name",
                sqlalchemy.String(length=100),
            ),
            sqlalchemy.Column("version", sqlalchemy.BigInteger(), nullable=False),
        ]

    @classmethod
    def get_special_columns(cls):
        sc = super().get_special_columns()
        sc["id"].extend(["backend_id", "tenant_id"])
        return sc


class RWKeyModel(RWMysqlModel):
    table_name = "key"

    class ValidationModel(RWValidationModelBase):
        login: str
        enabled: RWYesNoEnum
        password: str

    @staticmethod
    def get_db_columns():
        return [
            sqlalchemy.Column("login", sqlalchemy.String(length=250), nullable=False),
            sqlalchemy.Column(
                "password_hash", sqlalchemy.String(length=250), nullable=False
            ),
            sqlalchemy.Column("enabled", sqlalchemy.Enum("Y", "N"), nullable=False),
            sqlalchemy.Column("mfa_enabled", sqlalchemy.Enum("Y", "N"), nullable=False),
            sqlalchemy.Column("mfa_secret", sqlalchemy.String(length=250)),
        ]

    @classmethod
    def get_special_columns(cls):
        sc = super().get_special_columns()
        sc["encrypted"].append("mfa_secret")
        return sc

    @classmethod
    def _create_password_hash(cls, row):
        password = row.pop("password", None)
        if password:
            salt = os.urandom(32)
            kdf = Scrypt(
                salt=salt, length=32, n=2 ** 14, r=8, p=1, backend=default_backend()
            )
            b = kdf.derive(password.encode())
            row["password_hash"] = "%s.%s" % (salt.hex(), b.hex())
        return row

    @classmethod
    async def before_create(cls, db, row, skip_validation=False):
        row = await super().before_create(db, row, skip_validation)
        row = cls._create_password_hash(row)
        return row

    @classmethod
    async def before_update(cls, db, row, skip_validation=False):
        row = await super().before_update(db, row, skip_validation)
        row = cls._create_password_hash(row)
        return row

    @staticmethod
    def check_password(password_hash, password):
        try:
            salt, key = password_hash.split(".")
            kdf = Scrypt(
                salt=bytes.fromhex(salt),
                length=32,
                n=2 ** 14,
                r=8,
                p=1,
                backend=default_backend(),
            )
            kdf.verify(password.encode(), bytes.fromhex(key))
            return True

        except InvalidKey:
            return False

    @staticmethod
    def toggle_mfa(mfa_enabled, enabled):
        if enabled == RWYesNoEnum.YES:
            if mfa_enabled == RWYesNoEnum.YES:
                raise ValueError("MFA already enabled")

            return {"mfa_enabled": RWYesNoEnum.YES, "mfa_secret": pyotp.random_base32()}

        elif enabled == RWYesNoEnum.NO:
            if mfa_enabled == RWYesNoEnum.NO:
                raise ValueError("MFA already disabled")

            return {"mfa_enabled": RWYesNoEnum.NO, "mfa_secret": None}

    @staticmethod
    def check_otp(mfa_secret, otp_code):
        totp = pyotp.TOTP(mfa_secret)
        return totp.verify(otp_code)


class RWOAuth2ClientModel(RWMysqlModel):
    table_name = "oauth2_client"

    class ValidationModel(RWValidationModelBase):
        name: str
        secret: str
        grant_types: str
        response_types: str
        redirect_uris: str
        scope: Optional[str]

    @staticmethod
    def get_db_columns():
        return [
            sqlalchemy.Column("name", sqlalchemy.String(length=250), nullable=False),
            sqlalchemy.Column("secret", sqlalchemy.String(length=150), nullable=False),
            sqlalchemy.Column(
                "grant_types", sqlalchemy.String(length=150), nullable=False
            ),
            sqlalchemy.Column(
                "response_types", sqlalchemy.String(length=150), nullable=False
            ),
            sqlalchemy.Column(
                "redirect_uris", sqlalchemy.String(length=500), nullable=False
            ),
            sqlalchemy.Column("scope", sqlalchemy.String(length=150)),
        ]

    @classmethod
    def get_special_columns(cls):
        sc = super().get_special_columns()
        sc["encrypted"].append("secret")
        return sc


class RWOAuth2AuthorizationCodeModel(RWMysqlModel):
    table_name = "oauth2_authorization_code"

    class ValidationModel(RWValidationModelBase):
        code: str
        oauth2_client_id: str
        key_id: str
        redirect_uri: str
        response_type: str
        scope: Optional[str]
        auth_time: datetime
        expires_in: int
        nonce: Optional[str]
        code_challenge: str
        code_challenge_method: str

    @staticmethod
    def get_db_columns():
        return [
            sqlalchemy.Column("code", sqlalchemy.String(length=250), nullable=False),
            sqlalchemy.Column(
                "oauth2_client_id",
                sqlalchemy.Binary(16),
                sqlalchemy.ForeignKey("oauth2_client.id"),
                nullable=False,
            ),
            sqlalchemy.Column(
                "key_id",
                sqlalchemy.Binary(16),
                sqlalchemy.ForeignKey("key.id"),
                nullable=False,
            ),
            sqlalchemy.Column(
                "redirect_uri", sqlalchemy.String(length=150), nullable=False
            ),
            sqlalchemy.Column(
                "response_type", sqlalchemy.String(length=150), nullable=False
            ),
            sqlalchemy.Column("scope", sqlalchemy.String(length=150), nullable=False),
            sqlalchemy.Column("auth_time", sqlalchemy.DateTime(), nullable=False),
            sqlalchemy.Column("expires_in", sqlalchemy.Integer(), nullable=False),
            sqlalchemy.Column("nonce", sqlalchemy.String(length=150)),
            sqlalchemy.Column(
                "code_challenge", sqlalchemy.String(length=150), nullable=False
            ),
            sqlalchemy.Column(
                "code_challenge_method", sqlalchemy.String(length=10), nullable=False
            ),
        ]

    @classmethod
    def get_special_columns(cls):
        sc = super().get_special_columns()
        sc["id"].extend(["oauth2_client_id", "key_id"])
        return sc


class RWOAuth2TokenModel(RWMysqlModel):
    table_name = "oauth2_token"

    class ValidationModel(RWValidationModelBase):
        access_token: str
        refresh_token: Optional[str]
        scope: Optional[str]
        issued_at: datetime
        access_token_expires_in: int
        refresh_token_expires_in: int
        oauth2_client_id: str
        key_id: str
        token_type: str = "Bearer"
        revoked: RWYesNoEnum = RWYesNoEnum.NO

    @staticmethod
    def get_db_columns():
        return [
            sqlalchemy.Column(
                "access_token", sqlalchemy.String(length=250), nullable=False
            ),
            sqlalchemy.Column("refresh_token", sqlalchemy.String(length=250)),
            sqlalchemy.Column("scope", sqlalchemy.String(length=250)),
            sqlalchemy.Column("issued_at", sqlalchemy.DateTime(), nullable=False),
            sqlalchemy.Column(
                "access_token_expires_in", sqlalchemy.Integer(), nullable=False
            ),
            sqlalchemy.Column(
                "refresh_token_expires_in", sqlalchemy.Integer(), nullable=False
            ),
            sqlalchemy.Column(
                "oauth2_client_id",
                sqlalchemy.Binary(16),
                sqlalchemy.ForeignKey("oauth2_client.id"),
                nullable=False,
            ),
            sqlalchemy.Column(
                "key_id", sqlalchemy.Binary(16), sqlalchemy.ForeignKey("key.id")
            ),
            sqlalchemy.Column(
                "token_type", sqlalchemy.String(length=250), nullable=False
            ),
            sqlalchemy.Column(
                "revoked", sqlalchemy.Enum("Y", "N"), default="N", nullable=False
            ),
        ]

    @classmethod
    def get_special_columns(cls):
        sc = super().get_special_columns()
        sc["id"].extend(["oauth2_client_id", "key_id"])
        return sc


class RWAclRoleModel(RWMysqlModel):
    table_name = "acl_role"

    class ValidationModel(RWValidationModelBase):
        name: str

    @staticmethod
    def get_db_columns():
        return [
            sqlalchemy.Column("name", sqlalchemy.String(length=250), nullable=False),
        ]


class RWAclPermissionModel(RWMysqlModel):
    table_name = "acl_permission"

    class ValidationModel(RWValidationModelBase):
        name: str

    @staticmethod
    def get_db_columns():
        return [
            sqlalchemy.Column("name", sqlalchemy.String(length=250), nullable=False),
        ]


class RWAclRolePermissionLinkModel(RWMysqlModel):
    table_name = "acl_role_permission_link"

    class ValidationModel(RWValidationModelBase):
        acl_role_id: str
        acl_permission_id: str

    @staticmethod
    def get_db_columns():
        return [
            sqlalchemy.Column(
                "acl_role_id",
                sqlalchemy.Binary(16),
                sqlalchemy.ForeignKey("acl_role.id"),
                nullable=False,
            ),
            sqlalchemy.Column(
                "acl_permission_id",
                sqlalchemy.Binary(16),
                sqlalchemy.ForeignKey("acl_permission.id"),
                nullable=False,
            ),
        ]

    @classmethod
    def get_special_columns(cls):
        sc = super().get_special_columns()
        sc["id"].extend(["acl_role_id", "acl_permission_id"])
        return sc

    @classmethod
    def get_select_query(cls):
        fields = [cls.c.id]
        for c in RWAclRoleModel.c:
            fields.append(c.label(str(c).replace(".", "_")))
        for c in RWAclPermissionModel.c:
            fields.append(c.label(str(c).replace(".", "_")))

        return sqlalchemy.select(fields).select_from(cls.select().alias())


class RWAclRoleKeyLinkModel(RWMysqlModel):
    table_name = "acl_role_key_link"

    class ValidationModel(RWValidationModelBase):
        acl_role_id: str
        acl_key_id: str

    @staticmethod
    def get_db_columns():
        return [
            sqlalchemy.Column(
                "acl_role_id",
                sqlalchemy.Binary(16),
                sqlalchemy.ForeignKey("acl_role.id"),
                nullable=False,
            ),
            sqlalchemy.Column(
                "acl_key_id",
                sqlalchemy.Binary(16),
                sqlalchemy.ForeignKey("acl_key.id"),
                nullable=False,
            ),
        ]

    @classmethod
    def get_special_columns(cls):
        sc = super().get_special_columns()
        sc["id"].extend(["acl_role_id", "acl_key_id"])
        return sc



import asyncio
from starlette.applications import Starlette
from starlette.responses import HTMLResponse
from red_warden.config import RWConfig, Logger, Engines, Middlewares
from red_warden.datalayer import RWDataLayer, RWDataLayerException
from red_warden.responses import RWJSONResponse
from red_warden.helpers import exception_as_dict
from red_warden.mvc import RWStaticFilesController
from red_warden.datazones.RWManager import RWManager

__version__ = "0.4.2"
Logger.info("Initializing RedWarden %s" % RWConfig["RW_NAME"])


class RedWarden:
    frontend_exception_handlers = {
        401: None,
        403: None,
        404: None,
        500: None,
    }
    _app = None
    _routes = []

    @classmethod
    def add_datazones(cls, datazones):
        for datazone in datazones:
            RWDataLayer.add_datazone(
                datazone.__name__,
                datazone.models,
                datazone.controllers,
                datazone.multi_tenancy,
                datazone.schema_path,
            )

            # datazone_permissions = set()
            for controller in datazone.controllers:
                cls._routes.extend(controller.get_routes())

            #     for route in controller.get_routes():
            #         if isinstance(route, Mount):
            #             for _route in route.routes:
            #                 if _route.permissions:
            #                     datazone_permissions |= set(_route.permissions)
            #         else:
            #             if route.permissions:
            #                 datazone_permissions |= set(route.permissions)
            #
            # acl.add({datazone.__name__: datazone_permissions})

    @classmethod
    def add_static_files_controller(cls, project_dir, output_dir):
        ctrl = RWStaticFilesController(project_dir, output_dir)
        cls._routes.extend(ctrl.get_routes())

    # def add_graphql_controller(self, queries, mutations, prefix="/api"):
    #     self.add_controllers(
    #         RWGraphQLController(
    #             create_query_container(queries),
    #             create_mutation_container(mutations),
    #             prefix,
    #         )
    #     )

    @staticmethod
    def handle_loop_exception(loop, context):
        # context["message"] will always be there; but context["exception"] may not
        msg = context.get("exception", context["message"])
        Logger.error(f"Asyncio loop exception: {msg}")

    @classmethod
    async def _401_default_handler(cls, request, ex):
        if hasattr(request.state, "api") and request.state.api:
            return RWJSONResponse({"error": "Unauthorized"}, status_code=401)
        elif cls.frontend_exception_handlers[401]:
            return await cls.frontend_exception_handlers[401](request, ex)
        else:
            return HTMLResponse("<h1>UNAUTHORIZED :((</h1>", status_code=401)

    @classmethod
    async def _403_default_handler(cls, request, ex):
        if hasattr(request.state, "api") and request.state.api:
            return RWJSONResponse({"error": "Forbidden"}, status_code=403)
        elif cls.frontend_exception_handlers[403]:
            return await cls.frontend_exception_handlers[403](request, ex)
        else:
            return HTMLResponse("<h1>FORBIDDEN :((</h1>", status_code=403)

    @classmethod
    async def _404_default_handler(cls, request, ex):
        if hasattr(request.state, "api") and request.state.api:
            return RWJSONResponse({"error": "Not found"}, status_code=404)
        elif cls.frontend_exception_handlers[404]:
            return await cls.frontend_exception_handlers[404](request, ex)
        else:
            return HTMLResponse("<h1>NOT FOUND :((</h1>", status_code=404)

    @classmethod
    async def _500_default_handler(cls, request, ex):
        dict_ex = exception_as_dict(ex)
        if hasattr(request.state, "api") and request.state.api:
            return RWJSONResponse({"error": dict_ex}, status_code=500)
        elif cls.frontend_exception_handlers[500]:
            return await cls.frontend_exception_handlers[500](request, ex)
        else:
            return HTMLResponse("<pre>Error: %s</pre>" % str(dict_ex), status_code=500)

    @classmethod
    def start(cls):
        exception_handlers = {
            401: cls._401_default_handler,
            403: cls._403_default_handler,
            404: cls._404_default_handler,
            500: cls._500_default_handler,
        }

        cls.add_datazones([RWManager])

        # if RWConfig["RW_DEBUG"]:
        #     for r in cls._routes:
        #         if isinstance(r, Mount):
        #             for _r in r.routes:
        #                 Logger.info(_r.name)
        #         else:
        #             Logger.info(r.name)

        cls._app = Starlette(
            routes=cls._routes,
            middleware=Middlewares.get_all(),
            on_startup=[cls.on_startup],
            on_shutdown=[cls.on_shutdown],
            exception_handlers=exception_handlers,
        )

        asyncio.get_event_loop().set_exception_handler(cls.handle_loop_exception)
        return cls._app

    @classmethod
    async def on_startup(cls):
        Logger.info("Starting RedWarden %s" % RWConfig["RW_NAME"])

        await RWDataLayer.upgrade_manager_datapath_schema()

        for name, engine in Engines.get_all().items():
            asyncio.create_task(engine.run())

    @classmethod
    async def on_shutdown(cls):
        Logger.info("Terminating Red Warden...")
        for name, engine in Engines.get_all().items():
            await engine.stop()
